"""
model the Subaru optics system

This is a code modified from Rupert's original optics_propagate.py. This code adds more optics to the system,
as well as puts the AO, etc in order for Subaru.

Here, we will add the basic functionality of the Subaru Telescope, including the primary, secondary, and AO188.
The SCExAO system sits behind the AO188 instrument of Subaru, which is a 188-element AO system located at the
Nasmyth focus (IR) of the telescope. AO188 is a basic 4f-type optical system with a 188 element DM (not perfectly
represented here. Proper will only simulate 2D square DM's, so we use a 14x14 square) inserted in the middle of the
collimated beam. This routine simulates an idealized detector at the A0188 focus.

If you want to scale the intensity of the companion, you can't then normalize the intensity
(prop_define_entrance) since the normalization is per wavefront, so the companion will normalize to itself
rather than to the star. Doing it this way may cause confusion, but better than re-writing things in proper

AO188 uses laser guide-star technology. More info here:
https://subarutelescope.org/Introduction/instrument/AO188.html

A more detailed model of SCExAO will be modelled in a SCExAO_optics.py code. However, this routine is designed for
simple simulations that need to optimize runtime but still have relevance to the Subaru Telescope.

This script is meant to override any Subaru/SCExAO-specific parameters specified in the user's params.py
"""

import numpy as np
from inspect import getframeinfo, stack
import proper

from medis.params import ap, tp, sp
from medis.utils import dprint
import medis.optics as opx
import medis.aberrations as aber
import medis.adaptive as ao
import medis.atmosphere as atmos


#################################################################################################
#################################################################################################
#################################################################################################
# Defining Subaru parameters
# ----------------------------
# According to Iye-et.al.2004-Optical_Performance_of_Subaru:AstronSocJapan, the AO188 uses the IR-Cass secondary,
# but then feeds it to the IR Nasmyth f/13.6 focusing arrangement. So instead of simulating the full Subaru system,
# we can use the effective focal length at the Nasmyth focus, and simulate it as a single lens.
tp.d_nsmyth = 7.9716  # m pupil diameter
tp.fn_nsmyth = 13.612  # f# Nasmyth focus
tp.flen_nsmyth = tp.d_nsmyth * tp.fn_nsmyth  # m focal length
tp.dist_nsmyth_ao1 = tp.flen_nsmyth + 1.14  # m distance secondary to M1 of AO188 (hand-tuned, could update with
                                            # data from literature)

#  Below are the actual dimenstions of the Subaru telescope.
# --------------------------------
# tp.enterence_d = 8.2  # m diameter of primary
# tp.flen_primary = 15  # m focal length of primary
# tp.dist_pri_second = 12.652  # m distance primary -> secondary
# Secondary
tp.d_secondary = 1.265  # m diameter secondary, used for central obscuration
# tp.fn_secondary = 12.6

# Re-writing params terms in Subaru-units
# need this to accurately make atmospheric and aberration maps
tp.entrance_d = tp.d_nsmyth
tp.flen_primary = tp.flen_nsmyth

# Effective Primary Aberrations
# primary_aber_vals = {'a': [7.2e-17, 3e-17],  # power at low spatial frequencies (m4)
#                      'b': [0.8, 0.2],  # correlation length (b/2pi defines knee)
#                      'c': [3.1, 0.5],  #
#                      'a_amp': [0.05, 0.01]}
# ----------------------------
# AO188 OAP1
# Paramaters taken from "Design of the Subaru laser guide star adaptive optics module"
#  Makoto Watanabe et. al. SPIE doi: 10.1117/12.551032
tp.d_ao1 = 0.20  # m  diamater of AO1
tp.fl_ao1 = 1.201  # m  focal length OAP1
tp.dist_ao1_dm = 1.345  # m distance OAP1 to DM
# OAP1_aber_vals = {'a': [7.2e-17, 3e-17],  # power at low spatial frequencies (m4)
#                   'b': [0.8, 0.2],  # correlation length (b/2pi defines knee)
#                   'c': [3.1, 0.5],  #
#                   'a_amp': [0.05, 0.01]}

# ----------------------------
# AO188 OAP2
tp.dist_dm_ao2 = 2.511-tp.dist_ao1_dm  # m distance DM to OAP2
tp.d_ao2 = 0.2  # m  diamater of AO2
tp.fl_ao2 = 1.201  # m  focal length AO2
tp.dist_oap2_focus = 1.261
# OAP2_aber_vals = {'a': [7.2e-17, 3e-17],  # power at low spatial frequencies (m4)
#                   'b': [0.8, 0.2],  # correlation length (b/2pi defines knee)
#                   'c': [3.1, 0.5],  #
#                   'a_amp': [0.05, 0.01]}

tp.lens_params = [{'aber_vals': [7.2e-17, 0.8, 3.1],
                   'diam': tp.entrance_d,
                   'fl': tp.flen_nsmyth,
                   'dist': tp.dist_nsmyth_ao1,
                   'name': 'effective-primary'},

                  {'aber_vals': [7.2e-17, 0.8, 3.1],
                   'diam': tp.d_ao1,
                   'fl': tp.fl_ao1,
                   'dist': tp.dist_ao1_dm,
                   'name': 'ao188-OAP1'},

                  {'aber_vals': [7.2e-17, 0.8, 3.1],
                   'diam': tp.d_ao2,
                   'fl': tp.fl_ao2,
                   'dist': tp.dist_oap2_focus,
                   'name': 'ao188-OAP2'}
                    ]
#################################################################################################
#################################################################################################
#################################################################################################

def Subaru_frontend(empty_lamda, grid_size, PASSVALUE):
    """
    propagates instantaneous complex E-field thru Subaru from the primary through the AO188
        AO system in loop over wavelength range

    this function is called a 'prescription' by proper

    uses PyPROPER3 to generate the complex E-field at the source, then propagates it through atmosphere,
        then telescope, to the focal plane
    the AO simulator happens here
    this does not include the observation of the wavefront by the detector
    :returns spectral cube at instantaneous time in the focal_plane()
    """
    # print("Propagating Broadband Wavefront Through Subaru")

    # Initialize the Wavefront in Proper
    wfo = opx.Wavefronts()
    wfo.initialize_proper()

    # Atmosphere
    # atmos has only effect on phase delay, not intensity
    wfo.loop_collection(atmos.add_atmos, PASSVALUE['iter'], plane_name='atmosphere')

    # Defines aperture (baffle-before primary)
    # Obscurations (Secondary and Spiders)
    wfo.loop_collection(opx.add_obscurations, d_primary=tp.d_nsmyth, d_secondary=tp.d_secondary, legs_frac=0.05)
    wfo.loop_collection(proper.prop_circular_aperture,
                           **{'radius': tp.entrance_d / 2})  # clear inside, dark outside
    wfo.loop_collection(proper.prop_define_entrance, plane_name='entrance_pupil')  # normalizes abs intensity

    if ap.companion:
        # Must do this after all calls to prop_define_entrance
        wfo.loop_collection(opx.offset_companion)
        wfo.loop_collection(proper.prop_circular_aperture,
                               **{'radius': tp.entrance_d / 2})  # clear inside, dark outside

    # Test Sampling
    if sp.verbose:
        opx.check_sampling(PASSVALUE['iter'], wfo, "initial", getframeinfo(stack()[0][0]), units='mm')
    # Testing Primary Focus (instead of propagating to focal plane)
    # wfo.loop_collection(opx.prop_pass_lens, tp.flen_nsmyth, tp.flen_nsmyth)  # test only going to prime focus

    ########################################
    # Subaru Propagation
    #######################################
    # Effective Primary
    # CPA from Effective Primary
    wfo.loop_collection(aber.add_aber, step=PASSVALUE['iter'], lens_name='ao188-OAP1')
    # Zernike Aberrations- Low Order
    # wfo.loop_collection(aber.add_zern_ab, tp.zernike_orders, aber.randomize_zern_values(tp.zernike_orders))
    wfo.loop_collection(opx.prop_pass_lens, tp.flen_nsmyth, tp.dist_nsmyth_ao1)

    ########################################
    # AO188 Propagation
    ########################################
    # # AO188-OAP1
    wfo.loop_collection(aber.add_aber, step=PASSVALUE['iter'], lens_name='ao188-OAP1')
    wfo.loop_collection(opx.prop_pass_lens, tp.fl_ao1, tp.dist_ao1_dm)

    # AO System
    if tp.use_ao:
        WFS_map = ao.open_loop_wfs(wfo)
        wfo.loop_collection(ao.deformable_mirror, WFS_map, PASSVALUE['iter'], plane_name='woofer',
                            debug=sp.debug)  # don't use PASSVALUE['WFS_map'] here because open loop
    # ------------------------------------------------
    wfo.loop_collection(proper.prop_propagate, tp.dist_dm_ao2)

    # AO188-OAP2
    wfo.loop_collection(aber.add_aber, step=PASSVALUE['iter'], lens_name='ao188-OAP2')
    # wfo.loop_collection(aber.add_zern_ab, tp.zernike_orders, aber.randomize_zern_values(tp.zernike_orders)/2)
    wfo.loop_collection(opx.prop_pass_lens, tp.fl_ao2, tp.dist_oap2_focus)

    ########################################
    # Focal Plane
    # #######################################
    # Check Sampling in focal plane
    if sp.verbose:
        wfo.loop_collection(opx.check_sampling, PASSVALUE['iter'], "focal plane",
                            getframeinfo(stack()[0][0]), units='nm')

    # wfo.focal_plane fft-shifts wfo from Fourier Space (origin==lower left corner) to object space (origin==center)
    cpx_planes, sampling = wfo.focal_plane()

    print(f"Finished datacube at timestep = {PASSVALUE['iter']}")

    return cpx_planes, sampling




"""
model the end-to-end Subaru-SCExAO optics system

Here, we will add the basic functionality of the Subaru+SCExAO Telescope, including the primary, secondary, and AO188.
The SCExAO system sits behind the AO188 instrument of Subaru, which is a 188-element AO system located at the
Nasmyth focus (IR) of the telescope.

AO188 is a basic 4f-type optical system with a 188 element circular DM. Proper will only simulate 2D square DM's, so we
use a 14x14 square inserted in the middle of the collimated beam. We also artifically oversize the number of AO188
actuators to be a 16x16 array since we oversize the DM to the beam by one actuator on either side of the beam.

This script is meant to override any Subaru/SCExAO-specific parameters specified in the user's params.py
"""

import numpy as np
from inspect import getframeinfo, stack
import proper

from medis.params import iop, sp, ap, tp
from medis.utils import dprint
import medis.optics as opx
import medis.aberrations as aber
import medis.adaptive as ao
import medis.atmosphere as atmos
import medis.coronagraphy as cg

#################################################################################################
#################################################################################################
#################################################################################################
# iop.update_testname('SCExAO-dummy-save')

# Defining Subaru parameters
# ----------------------------
# According to Iye-et.al.2004-Optical_Performance_of_Subaru:AstronSocJapan, the AO188 uses the IR-Cass secondary,
# but then feeds it to the IR Nasmyth f/13.6 focusing arrangement. So instead of simulating the full Subaru system,
# we can use the effective focal length at the Nasmyth focus, and simulate it as a single lens.
tp.d_nsmyth = 7.9716  # m pupil diameter
tp.fn_nsmyth = 13.612  # f# Nasmyth focus
tp.flen_nsmyth = tp.d_nsmyth * tp.fn_nsmyth  # m focal length
tp.dist_nsmyth_ao1 = tp.flen_nsmyth + 1.14  # m distance secondary to M1 of AO188 (hand-tuned, could update with
                                            # data from literature)

#  Below are the actual dimenstions of the Subaru telescope.
# --------------------------------
# tp.enterence_d = 8.2  # m diameter of primary
# tp.flen_primary = 15  # m focal length of primary
# tp.dist_pri_second = 12.652  # m distance primary -> secondary
# Secondary
tp.d_secondary = 1.265  # m diameter secondary, used for central obscuration
# tp.fn_secondary = 12.6

# Re-writing params terms in Subaru-units
# need this to accurately make atmospheric and aberration maps
tp.entrance_d = tp.d_nsmyth
tp.flen_primary = tp.flen_nsmyth

# ----------------------------
# AO188 DM
tp.act_woofer = 15  # approximately a 188 DM (14*14=169) but then we include +2 pixels because the dm map is oversized
                    # by 2 pixels around the edge of the array

# ----------------------------
# AO188 OAP1
# Paramaters taken from "Design of the Subaru laser guide star adaptive optics module"
#  Makoto Watanabe et. al. SPIE doi: 10.1117/12.551032
tp.d_ao1 = 0.20  # m  diamater of AO1
tp.fl_ao1 = 1.201  # m  focal length OAP1
tp.dist_ao1_dm = 1.345  # m distance OAP1 to DM

# ----------------------------
# AO188 OAP2
tp.dist_dm_ao2 = 2.511-tp.dist_ao1_dm  # m distance DM to OAP2
tp.d_ao2 = 0.2  # m  diamater of AO2
tp.fl_ao2 = 1.201  # m  focal length AO2
tp.dist_oap2_focus = 1.261

# ------------------------------
# SCExAO
# These params aren't actually working, so just doing very basic, 4F optical systems until further notice

tp.d_tweeter = 0.051  # diameter of optics in SCExAO train are 2 inches=0.051 m
tp.act_tweeter = 50  # SCExAO actuators are 50x50=2500 actuators
tp.fl_SxOAPG = 0.255  # m focal length of Genera SCExAO lens (OAP1,3,4,5)
tp.fl_SxOAP2 = 0.519  # m focal length of SCExAO OAP 2
tp.d_SxOAPG = 0.051  # diameter of SCExAO OAP's
# tp.dist_cg_sl1 = tp.fl_SxOAPG + .000001  # m distance between AO188 focus and scexao lens1

tp.dist_SxOAP1_scexao = 0.1345  # m
tp.dist_scexao_sl2 = 0.2511 - tp.dist_SxOAP1_scexao  # m
tp.dist_sl2_focus = 0.1261  # m

tp.lens_params = [{'aber_vals': [7.2e-17, 0.8, 3.1],
                   'diam': tp.entrance_d,
                   'fl': tp.flen_nsmyth,
                   'dist': tp.dist_nsmyth_ao1,
                   'name': 'effective-primary'},

                  {'aber_vals': [7.2e-17, 0.8, 3.1],
                   'diam': tp.d_ao1,
                   'fl': tp.fl_ao1,
                   'dist': tp.dist_ao1_dm,
                   'name': 'ao188-OAP1'},

                  {'aber_vals': [7.2e-17, 0.8, 3.1],
                   'diam': tp.d_ao2,
                   'fl': tp.fl_ao2,
                   'dist': tp.dist_oap2_focus,
                   'name': 'ao188-OAP2'},

                 {'aber_vals': [7.2e-17, 0.8, 3.1],
                   'diam': tp.d_SxOAPG,
                   'fl': tp.fl_SxOAPG,
                   'dist': tp.fl_SxOAPG,
                   'name': 'SxOAPG'},

                 {'aber_vals': [7.2e-17, 0.8, 3.1],
                   'diam': tp.d_SxOAPG,
                   'fl': tp.fl_SxOAP2,
                   'dist': tp.fl_SxOAP2,
                   'name': 'SxOAP2'}
                 ]
# ------------------------------
# Coronagraph
tp.cg_type = 'Gaussian'
tp.cg_size = 2  # physical size or lambda/D size
tp.cg_size_units = "l/D"  # "m" or "l/D"
# tp.fl_cg_lens = 0.1021  # m
tp.fl_cg_lens = tp.fl_SxOAPG
tp.lyot_size = 0.9  # units are in fraction of surface blocked

#################################################################################################
#################################################################################################
#################################################################################################

def Subaru_SCExAO(empty_lamda, grid_size, PASSVALUE):
    """
    propagates instantaneous complex E-field thru Subaru from the primary through SCExAO

    this function is called a 'prescription' by proper

    uses PyPROPER3 to generate the complex E-field at the source, then propagates it through atmosphere,
        then telescope, to the focal plane
    the AO simulator happens here
    this does not include the observation of the wavefront by the detector
    :returns spectral cube at instantaneous time in the focal_plane()
    """
    # print("Propagating Broadband Wavefront Through Subaru")

    # Initialize the Wavefront in Proper
    wfo = opx.Wavefronts(sp.debug)
    wfo.initialize_proper()

    # Atmosphere
    # atmos has only effect on phase delay, not intensity
    wfo.loop_collection(atmos.add_atmos, PASSVALUE['iter'], plane_name='atmosphere')

    # Defines aperture (baffle-before primary)
    wfo.loop_collection(opx.add_obscurations, d_primary=tp.entrance_d, d_secondary=tp.d_secondary, legs_frac=0.05)
    wfo.loop_collection(proper.prop_circular_aperture,
                        **{'radius': tp.entrance_d / 2})  # clear inside, dark outside
    wfo.loop_collection(proper.prop_define_entrance, plane_name='entrance_pupil')  # normalizes abs intensity

    if ap.companion:
        # Must do this after all calls to prop_define_entrance
        wfo.loop_collection(opx.offset_companion)
        wfo.loop_collection(proper.prop_circular_aperture,
                            **{'radius': tp.entrance_d / 2})  # clear inside, dark outside

    # Test Sampling
    if sp.verbose:
        wfo.loop_collection(opx.check_sampling, PASSVALUE['iter'], "Telescope Aperture",
                            getframeinfo(stack()[0][0]), units='mm')
    # Testing Primary Focus (instead of propagating to focal plane)
    # wfo.loop_collection(opx.prop_pass_lens, tp.flen_nsmyth, tp.flen_nsmyth)  # test only going to prime focus

    ########################################
    # Subaru Propagation
    #######################################
    # Effective Primary
    # CPA from Effective Primary
    wfo.loop_collection(aber.add_aber, step=PASSVALUE['iter'], lens_name='effective-primary')  # high order
    wfo.loop_collection(aber.add_zern_ab, tp.zernike_orders, aber.randomize_zern_values(tp.zernike_orders))  # low order
    wfo.loop_collection(opx.prop_pass_lens, tp.flen_nsmyth, tp.dist_nsmyth_ao1)
    ########################################
    # AO188 Propagation
    ########################################
    # # AO188-OAP1
    wfo.loop_collection(aber.add_aber,  step=PASSVALUE['iter'], lens_name='ao188-OAP1')  # high order
    wfo.loop_collection(opx.prop_pass_lens, tp.fl_ao1, tp.dist_ao1_dm)

    # AO System
    if tp.use_ao:
        WFS_map = ao.open_loop_wfs(wfo)
        wfo.loop_collection(ao.deformable_mirror, WFS_map, PASSVALUE['iter'], apodize=False,
                            plane_name='woofer', debug=sp.verbose)  # don't use PASSVALUE['WFS_map'] here because open loop
    # ------------------------------------------------
    wfo.loop_collection(proper.prop_propagate, tp.dist_dm_ao2)

    # AO188-OAP2
    wfo.loop_collection(aber.add_aber, step=PASSVALUE['iter'], lens_name='ao188-OAP2')  # high order CPA
    wfo.loop_collection(aber.add_zern_ab, tp.zernike_orders, aber.randomize_zern_values(tp.zernike_orders)/2)  # low order CPA
    wfo.loop_collection(opx.prop_pass_lens, tp.fl_ao2, tp.dist_oap2_focus)

    ########################################
    # SCExAO
    # #######################################
    # SXExAO Reimaging 1
    wfo.loop_collection(aber.add_aber, step=PASSVALUE['iter'], lens_name='SxOAPG')  # high order CPA
    wfo.loop_collection(proper.prop_propagate, tp.fl_SxOAPG)  # from AO188 focus to S-OAP1
    wfo.loop_collection(opx.prop_pass_lens, tp.fl_SxOAPG, tp.fl_SxOAPG)  # from SxOAP1 to tweeter-DM
    #
    # AO System
    if tp.use_ao:
        WFS_map = ao.open_loop_wfs(wfo)
        wfo.loop_collection(ao.deformable_mirror, WFS_map, PASSVALUE['iter'], apodize=False,
                            plane_name='tweeter', debug=sp.verbose)
    # ------------------------------------------------
    wfo.loop_collection(proper.prop_propagate, tp.fl_SxOAPG)  # from tweeter-DM to OAP2

    # SXExAO Reimaging 2
    wfo.loop_collection(aber.add_aber, step=PASSVALUE['iter'], lens_name='SxOAP2')  # high order NCPA
    wfo.loop_collection(aber.add_zern_ab, tp.zernike_orders, aber.randomize_zern_values(tp.zernike_orders)/2)  # low order NCPA
    wfo.loop_collection(opx.prop_pass_lens, tp.fl_SxOAP2, tp.fl_SxOAP2, plane_name='post-DM-focus')  #tp.dist_sl2_focus

    # wfo.loop_collection(opx.check_sampling, PASSVALUE['iter'], "post-DM-focus",
    #                     getframeinfo(stack()[0][0]), units='nm')

    # Coronagraph
    # settings should be put into tp, and are not implicitly passed here
    wfo.loop_collection(cg.coronagraph, occulter_mode=tp.cg_type, plane_name='coronagraph')

    ########################################
    # Focal Plane
    # #######################################
    # Check Sampling in focal plane
    # wfo.focal_plane fft-shifts wfo from Fourier Space (origin==lower left corner) to object space (origin==center)
    cpx_planes, sampling = wfo.focal_plane()

    if sp.verbose:
        wfo.loop_collection(opx.check_sampling, PASSVALUE['iter'], "focal plane",
                            getframeinfo(stack()[0][0]), units='nm')
        # opx.check_sampling(PASSVALUE['iter'], wfo, "focal plane", getframeinfo(stack()[0][0]), units='arcsec')

    if sp.verbose:
        print(f"Finished datacube at timestep = {PASSVALUE['iter']}")

    return cpx_planes, sampling




"""
model the Hubble optics system

This is a code modified from Rupert's original optics_propagate.py.

This script is meant to override any Hubble-specific parameters specified in the user's params.py
"""

import numpy as np
import proper
from scipy.interpolate import interp1d

from medis.params import iop, ap, tp, sp
from medis.utils import dprint
import medis.optics as opx
import medis.aberrations as aber
import medis.adaptive as ao
import medis.atmosphere as atmos

#################################################################################################
#################################################################################################
#################################################################################################
#iop.update_testname(Hubble-basic-test1')

# Defining Hubble parameters
# ----------------------------
tp.entrance_d = 2.4  # m
tp.flen_primary = 5.52085  # m
tp.dist_pri_second = 4.907028205  # m distance primary -> secondary

# --------------------------------
# Secondary
tp.d_secondary = 0.396  # m diameter secondary, used for central obscuration
tp.flen_secondary = -0.6790325  # m focal length of secondary
tp.dist_second_focus = 6.3919974  # m distance secondary to M1 of AO188

# --------------------------------
# Wavelength
ap.wvl_range = np.array([450, 580]) / 1e9
ap.n_wvl_init = 3
ap.n_wvl_final = 6
sp.subplt_cols = 3

tp.obscure = True
tp.use_ao = False

#################################################################################################
#################################################################################################
#################################################################################################

def Hubble_frontend(empty_lamda, grid_size, PASSVALUE):
    """
    propagates instantaneous complex E-field tSubaru from the primary through the AO188
        AO system in loop over wavelength range

    this function is called a 'prescription' by proper

    uses PyPROPER3 to generate the complex E-field at the source, then propagates it through atmosphere,
        then telescope, to the focal plane
    the AO simulator happens here
    this does not include the observation of the wavefront by the detector
    :returns spectral cube at instantaneous time, sampling of the wavefront at final location

    """
    # print("Propagating Broadband Wavefront Through Hubble Telescope")

    # Getting Parameters-import statements weren't working--RD
    passpara = PASSVALUE['params']
    ap.__dict__ = passpara[0].__dict__
    tp.__dict__ = passpara[1].__dict__
    iop.__dict__ = passpara[2].__dict__
    sp.__dict__ = passpara[3].__dict__

    datacube = []

    # Initialize the Wavefront in Proper
    wfo = opx.Wavefronts()
    wfo.initialize_proper()

    ########################################
    # Hubble Propagation
    #######################################
    # Defines aperture (baffle-before primary)
    wfo.loop_collection(proper.prop_circular_aperture, **{'radius': tp.entrance_d / 2})  # clear inside, dark outside
    # Obscurations
    wfo.loop_collection(opx.add_obscurations, d_primary=tp.entrance_d, d_secondary=tp.d_secondary, legs_frac=0.01)
    wfo.loop_collection(proper.prop_define_entrance)  # normalizes the intensity

    # Test Sampling
    if PASSVALUE['iter'] == 1:
        initial_sampling = proper.prop_get_sampling(wfo.wf_collection[0,0])
        dprint(f"initial sampling is {initial_sampling:.4f}")

    # Primary
    # CPA from Effective Primary
    # aber.add_aber(wfo.wf_collection, tp.entrance_d, step=PASSVALUE['iter'], lens_name='primary')

    # wfo.loop_collection(opx.prop_pass_lens, tp.flen_primary, tp.flen_primary)
    wfo.loop_collection(opx.prop_pass_lens, tp.flen_primary, tp.dist_pri_second)

    # Secondary
    # aber.add_aber(wfo.wf_collection, tp.d_secondary, step=PASSVALUE['iter'], lens_name='second')
    # # Zernike Aberrations- Low Order
    # wfo.loop_collection(aber.add_zern_ab, tp.zernike_orders, tp.zernike_vals)
    wfo.loop_collection(opx.prop_pass_lens, tp.flen_secondary, tp.dist_second_focus)

    ########################################
    # Focal Plane
    # #######################################
    # Converting Array of Arrays (wfo) into 3D array
    #  wavefront array is now (number_wavelengths x number_astro_bodies x sp.grid_size x sp.grid_size)
    #  prop_end moves center of the wavefront from lower left corner (Fourier space) back to the center
    #    ^      also takes square modulus of complex values, so gives units as intensity not field
    shape = wfo.wf_collection.shape
    for iw in range(shape[0]):
        for io in range(shape[1]):
            if sp.maskd_size != sp.grid_size:
                wframes = np.zeros((sp.maskd_size, sp.maskd_size))
                (wframe, sampling) = proper.prop_end(wfo.wf_collection[iw, io], EXTRACT=np.int(sp.maskd_size))
            else:
                wframes = np.zeros((sp.grid_size, sp.grid_size))
                (wframe, sampling) = proper.prop_end(wfo.wf_collection[iw, io])  # Sampling returned by proper is in m
            wframes += wframe  # adds 2D wavefront from all astro_bodies together into single wavefront, per wavelength
        # dprint(f"sampling in focal plane at wavelength={iw} is {sampling} m")
        datacube.append(wframes)  # puts each wavlength's wavefront into an array
                                  # (number_wavelengths x sp.grid_size x sp.grid_size)

    datacube = np.array(datacube)
    datacube = np.roll(np.roll(datacube, tp.pix_shift[0], 1), tp.pix_shift[1],
                       2)  # cirshift array for off-axis observing
    datacube = np.abs(datacube)  # get intensity from datacube

    # Interpolating spectral cube from ap.n_wvl_init discreet wavelengths to ap.n_wvl_final
    if ap.interp_wvl and ap.n_wvl_init > 1 and ap.n_wvl_init < ap.n_wvl_final:
        wave_samps = np.linspace(0, 1, ap.n_wvl_init)
        f_out = interp1d(wave_samps, datacube, axis=0)
        new_heights = np.linspace(0, 1, ap.n_wvl_final)
        datacube = f_out(new_heights)

    print('Finished datacube at single timestep')

    return datacube, sampling




"""
coronagraphy.py
Kristina Davis
Nov 2019

This script adds functionality related to coronagraphy of the telescope system. I have decided to add a class
of occulter to the module to take care of different types of coronagraphs


"""
import os
import numpy as np
import proper
from astropy.io.fits import getdata, writeto
# import cv2
import warnings

from medis.params import tp, sp, ap, iop
import medis.optics as opx
from medis.utils import dprint


class Occulter():
    """
    produces an occulter of various modes

    valid modes:
    'Gaussian'
    'Solid'
    '8th_Order'

    most of this class is a modification of the chronograph routine found on the proper manual pg 85
    """
    def __init__(self, _mode):
        if _mode is 'Gaussian' or 'Solid' or '8th_Order':
            self.mode = _mode
        else:
            raise ValueError('please choose a valid mode: Gaussian, Solid, 8th_Order')

    def set_size(self, wf, size_in_lambda_d=0, size_in_m=0):
        """
        sets size of occulter in m, depending on if input is in units of lambda/D or in m

        :param wf: 2D wavefront
        :param size_in_lambda_d: desired occulter size in units of lambda/D
        :param size_in_m: desired occulter size in m
        :return: occulter size in m
        """
        lamda = proper.prop_get_wavelength(wf)
        dx_m = proper.prop_get_sampling(wf)
        dx_rad = proper.prop_get_sampling_radians(wf)

        if size_in_lambda_d is not 0:
            occrad_rad = size_in_lambda_d * lamda / tp.entrance_d  # occulter radius in radians
            self.size = occrad_rad * dx_m / dx_rad  # occulter radius in meters
        elif size_in_m is not 0:
            self.size = size_in_m
        else:
            raise ValueError('must set occulter size in either m or lambda/D units')

        # rescaling for focused system
        # if sp.focused_sys:
        #     self.size = wf.lamda / tp.entrance_d * ap.wvl_range[0] / wf.lamda

    def apply_occulter(self, wf):
        """
        applies the occulter by type spcified when class object was initiated

        :param wf: 2D wavefront
        :return:
        """
        # Code here pulled directly from Proper Manual pg 86
        if self.mode == "Gaussian":
            r = proper.prop_radius(wf)
            h = np.sqrt(-0.5 * self.size**2 / np.log(1 - np.sqrt(0.5)))
            gauss_spot = 1 - np.exp(-0.5 * (r/h)**2)
            # gauss_spot = shift(gauss_spot, shift=tp.occult_loc, mode='wrap')  # ???
            proper.prop_multiply(wf, gauss_spot)
        elif self.mode == "Solid":
            proper.prop_circular_obscuration(wf, self.size)
        elif self.mode == "8th_Order":
            proper.prop_8th_order_mask(wf, self.size, CIRCULAR=True)
        elif self.mode == 'Vortex':
            vortex = Vortex().occulter(wf)

    def apply_lyot(self, wf):
        """
        applies the appropriately sized Lyot stop depending on the coronagraph type

        :param wf: 2D wavefront
        :return:
        """
        if not hasattr(tp, 'lyot_size'):
            raise ValueError("must set tp.lyot_size in units fraction of the beam radius at the current surface")
        if self.mode is 'Gaussian':
            proper.prop_circular_aperture(wf, tp.lyot_size, NORM=True)
        elif self.mode is 'Solid':
            proper.prop_circular_aperture(wf, tp.lyot_size, NORM=True)
        elif self.mode is '8th_Order':
            proper.prop_circular_aperture(wf, tp.lyot_size, NORM=True)
        elif self.mode == "Vortex":
            vortex = Vortex().lyotstop(wf,True)

def coronagraph(wf, occulter_mode=None):
    """
    propagates a wavefront through a coronagraph system

    We implicitly assume here that the system begins at the location of the occulting mask. The optical system leading
    up to this point must be defined in the telescope perscription. This coronagraph routine also ends at the image
    plane after the Lyot stop, if used. If no Lyot stop is used, the reimaging optics just pass the pupil image
    unaffected.

    this function is modified from the Proper Manual: Stellar Coronagraph example found on pg 85

    :param wf: a single wavefront of complex data shape=(sp.grid_size, sp.grid_size)
    :param occulter_type string defining the occulter type. Accepted types are "Gaussian", "Solid", and "8th_Order"
    :return: noting is returned but wfo has been modified
    """
    if occulter_mode is None:
        pass
    else:
        # Initialize Occulter based on Mode
        occ = Occulter(occulter_mode)
        if not hasattr(tp, 'cg_size'):
            raise ValueError("must set tp.cg_size and tp.cg_size_units to use coronagraph() in coronography.py")
        elif tp.cg_size_units == "m":
            occ.set_size(wf, size_in_m=tp.cg_size)
        else:
            occ.set_size(wf, size_in_lambda_d=tp.cg_size)

        # Apply Occulter
        occ.apply_occulter(wf)  # size saved to self in class, so don't need to pass it

        proper.prop_propagate(wf, tp.fl_cg_lens)  # propagate to coronagraph pupil reimaging lens from occulter
        opx.prop_pass_lens(wf, tp.fl_cg_lens, tp.fl_cg_lens)  # go to the middle of the 2-lens system

        # Apply Lyot
        occ.apply_lyot(wf)

        proper.prop_propagate(wf, tp.fl_cg_lens)  # propagate to reimaging lens from lyot stop
        opx.prop_pass_lens(wf, tp.fl_cg_lens, tp.fl_cg_lens)  # go to the next image plane


def apodization(wf):
    raise NotImplementedError


class Vortex():
    # https://github.com/vortex-exoplanet/HEEPS/tree/master/heeps
    def __int__(self):
        warnings.warn("Using vector vortex code quickly adapted from METIS. Still largely unverified")

    def occulter(self, wf):

        n = int(proper.prop_get_gridsize(wf))
        ofst = 0  # no offset
        ramp_sign = 1  # sign of charge is positive
        ramp_oversamp = 11.  # vortex is oversampled for a better discretization

        # f_lens = tp.f_lens #conf['F_LENS']
        # diam = tp.diam#conf['DIAM']
        charge = 2#conf['CHARGE']
        pixelsize = 5#conf['PIXEL_SCALE']
        Debug_print = False#conf['DEBUG_PRINT']

        coron_temp = os.path.join(iop.testdir, 'coron_maps/')
        if not os.path.exists(coron_temp):
            os.mkdir(coron_temp)

        if charge != 0:
            wavelength = proper.prop_get_wavelength(wf)
            gridsize = proper.prop_get_gridsize(wf)
            beam_ratio = pixelsize * 4.85e-9 / (wavelength / tp.entrance_d)
            # dprint((wavelength,gridsize,beam_ratio))
            calib = str(charge) + str('_') + str(int(beam_ratio * 100)) + str('_') + str(gridsize)
            my_file = str(coron_temp + 'zz_perf_' + calib + '_r.fits')

            if (os.path.isfile(my_file) == True):
                if (Debug_print == True):
                    print ("Charge ", charge)
                vvc = self.readfield(coron_temp, 'zz_vvc_' + calib)  # read the theoretical vortex field
                vvc = proper.prop_shift_center(vvc)
                scale_psf = wf._wfarr[0, 0]
                psf_num = self.readfield(coron_temp, 'zz_psf_' + calib)  # read the pre-vortex field
                psf0 = psf_num[0, 0]
                psf_num = psf_num / psf0 * scale_psf
                perf_num = self.readfield(coron_temp, 'zz_perf_' + calib)  # read the perfect-result vortex field
                perf_num = perf_num / psf0 * scale_psf
                wf._wfarr = (
                             wf._wfarr - psf_num) * vvc + perf_num  # the wavefront takes into account the real pupil with the perfect-result vortex field

            else:  # CAL==1: # create the vortex for a perfectly circular pupil
                if (Debug_print == True):
                    print ("Charge ", charge)

                f_lens = 200.0 * tp.entrance_d
                wf1 = proper.prop_begin(tp.entrance_d, wavelength, gridsize, beam_ratio)
                proper.prop_circular_aperture(wf1, tp.entrance_d / 2)
                proper.prop_define_entrance(wf1)
                proper.prop_propagate(wf1, f_lens, 'inizio')  # propagate wavefront
                proper.prop_lens(wf1, f_lens, 'focusing lens vortex')  # propagate through a lens
                proper.prop_propagate(wf1, f_lens, 'VC')  # propagate wavefront

                self.writefield(coron_temp, 'zz_psf_' + calib, wf1.wfarr)  # write the pre-vortex field
                nramp = int(n * ramp_oversamp)  # oversamp
                # create the vortex by creating a matrix (theta) representing the ramp (created by atan 2 gradually varying matrix, x and y)
                y1 = np.ones((nramp,), dtype=np.int)
                y2 = np.arange(0, nramp, 1.) - (nramp / 2) - int(ramp_oversamp) / 2
                y = np.outer(y2, y1)
                x = np.transpose(y)
                theta = np.arctan2(y, x)
                x = 0
                y = 0
                vvc_tmp = np.exp(1j * (ofst + ramp_sign * charge * theta))
                theta = 0
                vvc_real_resampled = cv2.resize(vvc_tmp.real, (0, 0), fx=1 / ramp_oversamp, fy=1 / ramp_oversamp,
                                                interpolation=cv2.INTER_LINEAR)  # scale the pupil to the pupil size of the simualtions
                vvc_imag_resampled = cv2.resize(vvc_tmp.imag, (0, 0), fx=1 / ramp_oversamp, fy=1 / ramp_oversamp,
                                                interpolation=cv2.INTER_LINEAR)  # scale the pupil to the pupil size of the simualtions
                vvc = np.array(vvc_real_resampled, dtype=complex)
                vvc.imag = vvc_imag_resampled
                vvcphase = np.arctan2(vvc.imag, vvc.real)  # create the vortex phase
                vvc_complex = np.array(np.zeros((n, n)), dtype=complex)
                vvc_complex.imag = vvcphase
                vvc = np.exp(vvc_complex)
                vvc_tmp = 0.
                self.writefield(coron_temp, 'zz_vvc_' + calib, vvc)  # write the theoretical vortex field

                proper.prop_multiply(wf1, vvc)
                proper.prop_propagate(wf1, f_lens, 'OAP2')
                proper.prop_lens(wf1, f_lens)
                proper.prop_propagate(wf1, f_lens, 'forward to Lyot Stop')
                proper.prop_circular_obscuration(wf1, 1., NORM=True)  # null the amplitude iside the Lyot Stop
                proper.prop_propagate(wf1, -f_lens)  # back-propagation
                proper.prop_lens(wf1, -f_lens)
                proper.prop_propagate(wf1, -f_lens)
                self.writefield(coron_temp, 'zz_perf_' + calib, wf1.wfarr)  # write the perfect-result vortex field

                vvc = self.readfield(coron_temp, 'zz_vvc_' + calib)
                vvc = proper.prop_shift_center(vvc)
                scale_psf = wf._wfarr[0, 0]
                psf_num = self.readfield(coron_temp, 'zz_psf_' + calib)  # read the pre-vortex field
                psf0 = psf_num[0, 0]
                psf_num = psf_num / psf0 * scale_psf
                perf_num = self.readfield(coron_temp, 'zz_perf_' + calib)  # read the perfect-result vortex field
                perf_num = perf_num / psf0 * scale_psf
                wf._wfarr = (
                             wf._wfarr - psf_num) * vvc + perf_num  # the wavefront takes into account the real pupil with the perfect-result vortex field

        return wf

    def lyotstop(self, wf,  RAVC=None, APP=None, get_pupil='no', dnpup=50):
        """Add a Lyot stop, or an APP."""

        # load parameters
        npupil = 1#conf['NPUPIL']
        pad = int((210 - npupil) / 2)

        # get LS misalignments
        LS_misalignment = (np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0]) * npupil).astype(int)
        dx_amp, dy_amp, dz_amp = LS_misalignment[0:3]
        dx_phase, dy_phase, dz_phase = LS_misalignment[3:6]

        # case 1: Lyot stop (no APP)
        if APP is not True:

            # Lyot stop parameters: R_out, dR_in, spi_width
            # outer radius (absolute %), inner radius (relative %), spider width (m)
            (R_out, dR_in, spi_width) = [0.98, 0.03, 0]

            # Lyot stop inner radius at least as large as obstruction radius
            R_in = 0.15

            # case of a ring apodizer
            if RAVC is True:
                # define the apodizer transmission and apodizer radius [Mawet2013]
                # apodizer radius at least as large as obstruction radius
                T_ravc = 1 - (R_in ** 2 + R_in * np.sqrt(R_in ** 2 + 8)) / 4
                R_in /= np.sqrt(1 - T_ravc)

            # oversize Lyot stop inner radius
            R_in += dR_in

            # create Lyot stop
            proper.prop_circular_aperture(wf, R_out, dx_amp, dy_amp, NORM=True)
            if R_in > 0:
                proper.prop_circular_obscuration(wf, R_in, dx_amp, dy_amp, NORM=True)
            if spi_width > 0:
                for angle in [10]:
                    proper.prop_rectangular_obscuration(wf, 0.05 * 8, 8 * 1.3, ROTATION=20)
                    proper.prop_rectangular_obscuration(wf, 8 * 1.3, 0.05 * 8, ROTATION=20)
                    # proper.prop_rectangular_obscuration(wf, spi_width, 2 * 8, \
                    #                                     dx_amp, dy_amp, ROTATION=angle)

        # case 2: APP (no Lyot stop)
        else:
            # get amplitude and phase files
            APP_amp_file = os.path.join(conf['INPUT_DIR'], conf['APP_AMP_FILE'])
            APP_phase_file = os.path.join(conf['INPUT_DIR'], conf['APP_PHASE_FILE'])
            # get amplitude and phase data
            APP_amp = getdata(APP_amp_file) if os.path.isfile(APP_amp_file) \
                else np.ones((npupil, npupil))
            APP_phase = getdata(APP_phase_file) if os.path.isfile(APP_phase_file) \
                else np.zeros((npupil, npupil))
            # resize to npupil
            APP_amp = resize(APP_amp, (npupil, npupil), preserve_range=True, mode='reflect')
            APP_phase = resize(APP_phase, (npupil, npupil), preserve_range=True, mode='reflect')
            # pad with zeros to match PROPER gridsize
            APP_amp = np.pad(APP_amp, [(pad + 1 + dx_amp, pad - dx_amp), \
                                       (pad + 1 + dy_amp, pad - dy_amp)], mode='constant')
            APP_phase = np.pad(APP_phase, [(pad + 1 + dx_phase, pad - dx_phase), \
                                           (pad + 1 + dy_phase, pad - dy_phase)], mode='constant')
            # multiply the loaded APP
            proper.prop_multiply(wf, APP_amp * np.exp(1j * APP_phase))

        # get the pupil amplitude or phase for output
        if get_pupil.lower() in 'amplitude':
            return wf, proper.prop_get_amplitude(wf)[pad + 1 - dnpup:-pad + dnpup, pad + 1 - dnpup:-pad + dnpup]
        elif get_pupil.lower() in 'phase':
            return wf, proper.prop_get_phase(wf)[pad + 1 - dnpup:-pad + dnpup, pad + 1 - dnpup:-pad + dnpup]
        else:
            return wf

    def writefield(self, path, filename, field):
        writeto(path + filename + '_r.fits', field.real, header=None, overwrite=True)
        writeto(path + filename + '_i.fits', field.imag, header=None, overwrite=True)

    def readfield(self, path, filename):

        try:
            data_r, hdr = getdata(path + filename + '_r.fits', header=True)
            data_i = getdata(path + filename + '_i.fits')
        except:
            dprint('FileNotFoundError. Waiting...')
            import time
            time.sleep(10)
            data_r, hdr = getdata(path + filename + '_r.fits', header=True)
            data_i = getdata(path + filename + '_i.fits')

        field = np.array(data_r, dtype=complex)
        field.imag = data_i

        return (field)



"""
CDI.py
Kristina Davis
8/12/19

This module is used to set the CDI parameters for mini-medis.
"""
import numpy as np
import warnings
from matplotlib import pyplot as plt
from matplotlib.colors import LogNorm, SymLogNorm
import copy

from medis.params import tp, sp
from medis.utils import dprint

class CDIOut():
    '''Stole this idea from falco. Just passes an object you can slap stuff onto'''
    pass


class CDI_params():
    """
    contains the parameters of the CDI probes and phase sequence to apply to the DM
    """
    def __init__(self):
        # General
        self.use_cdi = False
        self.show_probe = False  # False , flag to plot phase probe or not
        self.which_DM = ''  # plane_name parameter of the DM to apply CDI probe to (must be a valid name for
        # the telescope sim you are running, eg 'tweeter'

        # Probe Dimensions (extent in pupil plane coordinates)
        self.probe_amp = 2e-6  # [m] probe amplitude, scale should be in units of actuator height limits
        self.probe_w = 10  # [actuator coordinates] width of the probe
        self.probe_h = 30  # [actuator coordinates] height of the probe
        self.probe_shift = (15, 15)  # [actuator coordinates] center position of the probe (should move off-center to
        # avoid coronagraph)
        self.probe_spacing = 10  # distance from the focal plane center to edge of the rectangular probed region

        # Phase Sequence of Probes
        self.phs_intervals = np.pi / 2  # [rad] phase interval over [0, 2pi]
        self.phase_integration_time = 0.01  # [s]  How long in sec to apply each probe in the sequence
        self.null_time = 0  # [s]  time between repeating probe cycles (data to be nulled using probe info)

    def __iter__(self):
        for attr, value in self.__dict__.items():
            yield attr, value

    def __name__(self):
        return self.__str__().split(' ')[0].split('.')[-1]

    def gen_phaseseries(self):
        """
        generate an array of phases per timestep for the CDI algorithm

        phase_series is used to populate cdi.phase_series, which may be longer than cdi.phase_cycle if multiple cycles
        are run, or probes may last for longer than one single timestep

        currently, I assume the timestream is not that long. Should only use this for short timestreams, and use a more
        efficient code for long simulations (scale of minutes or more)

        :return: phase_series  array of phases of CDI probes to apply to DM
        """
        self.phase_series = np.zeros(sp.numframes) * np.nan

        if self.use_cdi:
            # Repeating Probe Phases for Integration time
            self.phase_cycle = np.arange(0, 2 * np.pi, self.phs_intervals)  # FYI not inclusive of 2pi endpoint
            self.n_probes = len(self.phase_cycle)  # number of phase probes
            if self.n_probes % 2 != 0:
                raise ValueError(f"must have even number of phase probes\n\tchange cdi.phs_intervals")

            if self.phase_integration_time > sp.sample_time:
                phase_hold = self.phase_integration_time / sp.sample_time
                phase_1cycle = np.repeat(self.phase_cycle, phase_hold)
            elif self.phase_integration_time == sp.sample_time:
                phase_1cycle = self.phase_cycle
            else:
                raise ValueError(f"Cannot have CDI phase probe integration time less than sp.sample_time")

            # Repeating Cycle of Phase Probes for Simulation Duration
            full_simulation_time = sp.numframes * sp.sample_time
            time_for_one_cycle = len(phase_1cycle) * self.phase_integration_time + self.null_time

            if time_for_one_cycle > full_simulation_time and self.phase_integration_time >= sp.sample_time:
                warnings.warn(f"\nLength of one full CDI probe cycle (including nulling) exceeds the "
                              f"full simulation time \n"
                              f"not all phases will be used\n"
                              f"phase reconstruction will be incomplete")
                self.phase_series = phase_1cycle[0:sp.numframes]
            elif time_for_one_cycle <= full_simulation_time and self.n_probes < sp.numframes:
                print(f"\nCDI Params\n\tThere will be {sp.numframes - self.n_probes} "
                      f"nulling steps after timestep {self.n_probes}")
                self.phase_series[0:self.n_probes] = phase_1cycle
            else:
                warnings.warn(f"Haven't run into  CDI phase situation like this yet")
                raise NotImplementedError

        return self.phase_series

    def init_cout(self, nact):
        self.cout = CDIOut()
        self.cout.nact = nact
        self.cout.grid_size = sp.grid_size
        self.cout.beam_ratio = sp.beam_ratio
        self.cout.DM_probe_series = np.zeros((self.n_probes, nact, nact))

    def save_probe(self, ix, probe):
        self.cout.DM_probe_series[ix] = probe

    def save_tseries(self, ts):
        """saves output of medis fields as 2D intensity images for CDI postprocessing"""
        self.cout.tseries = ts

    def save_cout_to_disk(self, plot=False):


        # Fig
        if plot:
            if self.n_probes >= 4:
                nrows = 2
                ncols = self.n_probes//2
                figheight = 5
            else:
                nrows = 1
                ncols = self.n_probes
                figheight = 12

            fig, subplot = plt.subplots(nrows, ncols, figsize=(12, figheight))
            fig.subplots_adjust(wspace=0.5)

            fig.suptitle('Probe Series')

            for ax, ix in zip(subplot.flatten(), range(self.n_probes)):
                # im = ax.imshow(self.DM_probe_series[ix], interpolation='none', origin='lower')
                im = ax.imshow(self.cout.probe_series[ix], interpolation='none', origin='lower')

                ax.set_title(f"Probe " + r'$\theta$=' + f'{self.DM_phase_series[ix]/np.pi:.2f}' + r'$\pi$')

            cb = fig.colorbar(im)  #
            cb.set_label('um')

            plt.show()

# Sneakily Instantiating Class Objects here
cdi = CDI_params()


def config_probe(theta, nact, iw=0, ib=0):
    """
    create a probe shape to apply to the DM for CDI processing

    The probe applied to the DM to achieve CDI is that originally proposed in Giv'on et al 2011, doi: 10.1117/12.895117;
    and was used with proper in Matthews et al 2018, doi:  10.1117/1.JATIS.3.4.045001.

    The pupil coordinates used in those equations relate to the sampling of the pupil (plane of the DM). However, in
    Proper, the prop_dm uses a DM map that is the size of (n_ao_act, n_ao_act). This map was resampled from its
    original size to the actuator spacing, and the spacing in units of [m] is supplied as a keyword to the prop_dm
    function during the call. All that is to say, we apply the CDI probe using the coordinates of the DM actuators,
    and supply the probe height as an additive height to the DM map, which is passed to the prop_dm function.

    :param theta: phase of the probe
    :param nact: number of actuators in the mirror, should change if 'woofer' or 'tweeter'
    :param iw: index of wavelength number in ap.wvl_range
    :return: height of phase probes to add to the DM map in adaptive.py
    """
    x = np.linspace(-1/2-cdi.probe_shift[0]/nact, 1/2-cdi.probe_shift[0]/nact, nact)
    y = np.linspace(-1/2-cdi.probe_shift[1]/nact, 1/2-cdi.probe_shift[1]/nact, nact)
    X,Y = np.meshgrid(x, y)

    probe = cdi.probe_amp * np.sinc(cdi.probe_w * X) * np.sinc(cdi.probe_h * Y) \
            * np.sin(2*np.pi*cdi.probe_spacing*X + theta)

    # Testing FF propagation
    if sp.verbose and iw == 0 and ib == 0:  # and theta == cdi.phase_series[0]
        probe_ft = (1/np.sqrt(2*np.pi)) * np.fft.fftshift(np.fft.fft2(np.fft.ifftshift(probe)))

        # fig, ax = plt.subplots(1, 3, figsize=(12, 5))
        # fig.subplots_adjust(wspace=0.5)
        # ax1, ax2, ax3 = ax.flatten()
        #
        # fig.suptitle(f"spacing={cdi.probe_spacing}, Dimensions {cdi.probe_w}x{cdi.probe_h} "
        #              f"\nProbe Amp = {cdi.probe_amp}, " + r'$\theta$' + f"={theta/np.pi:.3f}" + r'$\pi$' + '\n')
        #
        # im1 = ax1.imshow(probe, interpolation='none', origin='lower')
        # ax1.set_title(f"Probe on DM \n(dm coordinates)")
        # cb = fig.colorbar(im1, ax=ax1)
        #
        # im2 = ax2.imshow(np.sqrt(probe_ft.imag ** 2 + probe_ft.real ** 2), interpolation='none', origin='lower')
        # ax2.set_title("Focal Plane Amplitude")
        # cb = fig.colorbar(im2, ax=ax2)
        #
        # ax3.imshow(np.arctan2(probe_ft.imag, probe_ft.real), interpolation='none', origin='lower', cmap='hsv')
        # ax3.set_title("Focal Plane Phase")

        # plt.show()

        # Fig 2
        fig, ax = plt.subplots(1, 3, figsize=(12, 5))
        fig.subplots_adjust(wspace=0.5)
        ax1, ax2, ax3 = ax.flatten()
        fig.suptitle(f'Real & Imaginary Probe Response in Focal Plane\n'
                     f' '+r'$\theta$'+f'={theta/np.pi:.3f}'+r'$\pi$'+f', n_actuators = {nact}\n')

        im1 = ax1.imshow(probe, interpolation='none', origin='lower')
        ax1.set_title(f"Probe on DM \n(dm coordinates)")
        cb = fig.colorbar(im1, ax=ax1)

        im2 = ax2.imshow(probe_ft.real, interpolation='none', origin='lower')
        ax2.set_title(f"Real FT of Probe")

        im3 = ax3.imshow(probe_ft.imag, interpolation='none', origin='lower')
        ax3.set_title(f"Imag FT of Probe")

        plt.show()

    # Saving Probe in the series
    if iw == 0 and ib == 0:
        ip = np.argwhere(cdi.phase_series == theta)
        cdi.save_probe(ip[0,0], probe)
        cdi.nact = nact

    return probe


def cdi_postprocess(fp_seq, sampling, plot=False):
    """
    this is the function that accepts the timeseries of intensity images from the simuation and returns the processed
    single image. This function calculates the speckle amplitude phase, and then corrects for it to create the dark
    hole over the specified region of the image.

    :param fp_seq: timestream of 2D images (intensity only) from the focal plane complex field
    :param sampling: focal plane sampling
    :return:
    """
    n_pairs = cdi.n_probes//2  # number of deltas (probe differentials)
    n_nulls = sp.numframes - cdi.n_probes
    delta = np.zeros((n_pairs, sp.grid_size, sp.grid_size), dtype=float)
    absDeltaP = np.zeros((n_pairs, sp.grid_size, sp.grid_size), dtype=float)
    phsDeltaP = np.zeros((n_pairs, sp.grid_size, sp.grid_size), dtype=float)
    Epupil = np.zeros((n_nulls, sp.grid_size, sp.grid_size), dtype=float)

    fp_seq = np.sum(fp_seq, axis=1)  # sum over wavelength

    for ix in range(n_pairs):
        # Compute deltas
        delta[ix] = np.copy(fp_seq[ix] - fp_seq[ix+n_pairs])

        for xn in range(n_nulls):
            dprint(f'for computing reDeltaP: xn={xn}')
            # Real Part of deltaP
            # absDeltaP[xn] = np.sqrt((fp_seq[ix] + fp_seq[ix+n_pairs])/2 - fp_seq[xn])
            # probe_ft = (1/np.sqrt(2*np.pi)) * np.fft.fftshift(np.fft.fft2(np.fft.ifftshift(cdi.cout.DM_probe_series[ix])))
            # phsDeltaP[xn] = np.arctan2(probe_ft.imag, probe_ft.real)

            # Least Squares Solution
            # Epupil[xn] = np.linalg.solve((-ImDeltaP[xn] + ReDeltaP[xn], delta[ix]))
            # duVecNby1 = -dmfac * np.linalg.solve((10 ** log10reg * np.diag(cvar.EyeGstarGdiag) + cvar.GstarG_wsum),
            #                                      cvar.RealGstarEab_wsum)
            # duVec = duVecNby1.reshape((-1,))

            # Fig 2
    if plot:
        fig, subplot = plt.subplots(1, n_pairs, figsize=(14,5))
        fig.subplots_adjust(wspace=0.5, right=0.85)

        fig.suptitle('Deltas for CDI Probes')

        for ax, ix in zip(subplot.flatten(), range(n_pairs)):
            im = ax.imshow(delta[ix]*1e6, interpolation='none', origin='lower',
                           norm=SymLogNorm(linthresh=1e-2),
                           vmin=-1, vmax=1) #, norm=SymLogNorm(linthresh=1e-5))
            ax.set_title(f"Diff Probe\n" + r'$\theta$' + f'={cdi.phase_series[ix]/np.pi:.3f}' +
                         r'$\pi$ -$\theta$' + f'={cdi.phase_series[ix+n_pairs]/np.pi:.3f}' + r'$\pi$')

        cax = fig.add_axes([0.9, 0.2, 0.03, 0.6])  # Add axes for colorbar @ position [left,bottom,width,height]
        cb = fig.colorbar(im, orientation='vertical', cax=cax)  #
        cb.set_label('Intensity')

        plt.show()


if __name__ == '__main__':
    dprint(f"Testing CDI probe")
    cdi.use_cdi = True; cdishow_probe = True

    cdi.probe_amp = 2e-6  # [m] probe amplitude, scale should be in units of actuator height limits
    cdi.probe_w = 10  # [actuator coordinates] width of the probe
    cdi.probe_h = 30  # [actuator coordinates] height of the probe
    cdi.probe_shift = [5, 5]  # [actuator coordinates] center position of the probe
    cdi.probe_spacing = 15

    tp.act_tweeter = 49

    sp.numframes = 10
    cdi.gen_phaseseries()
    cdi.init_probes(tp.act_tweeter)
    # cdiphase_series = [-1*np.pi/4]
    config_probe(cdi.phase_series[0], tp.act_tweeter)  #



"""
This is the main configuration file. It contains default global variables (as in they are read in by the relevant
modules) that define the parameters of the whole telescope system. These parameters can be redefined at the beginning
of the example module the user is running

Unless otherwise specified, units shall be given in:
distance: meters
time: seconds

TODO
    * Add all possible initial SP attributes here so the user knows what the possible options are. Consider adding other param types too
    * add user_params.py too so we don't have to keep changing rootdir and datadir when pushing?
"""

import numpy as np
import os
import proper
from pathlib import Path
import warnings

class IO_params:
    """
    Define file tree/structure to import and save data

    telescope.gridsize and telescope.beam_ratio affect the sampling of the simulation, so are included in filenames
    of files genereated based on the sampling parameters, and simulation_params.numframes affects how many files need
    to be generated, so are also included in filenames
    """

    def __init__(self, datadir=f'{str(Path.home())}/MKIDSim/', testname='example1'):
        # you can update the datadir and testname with iop.update_testname('your_testname') and then run iop.mkdir()
        self.datadir = datadir
        self.prescriptions_root = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
                                               'simulations')  # top level of where prescriptions found

        # Unprocessed Science Data
        self.testname = testname  # set this up in the definition line, but can update it with iop.update_testname('newname')
        self.testdir = os.path.join(self.datadir,
                                  self.testname)  # Save results in new sub-directory
        self.params_logs = os.path.join(self.testdir, 'params.pkl')

        self.fields = os.path.join(self.testdir, 'fields.h5')  # a x/y/t/w cube of data
        self.photonlist = os.path.join(self.testdir, 'photonlist.h5')  # a photon table with 4 columns
        self.camera = os.path.join(self.testdir, 'camera.pkl')  # MKIDS.Camera instance save state
        self.telescope = os.path.join(self.testdir, 'telescope.pkl')  # a telecope.Telescope instance save state

        self.atmosroot = 'atmos'
        atmosdir = "gridsz{}_bmratio{}_tsteps{}"  # fill in variable names later
        self.atmosdir = os.path.join(self.testdir, self.atmosroot, atmosdir)  # full path to FITS files

        # Aberration Metadata
        self.aberroot = 'aberrations'
        aberdir = "gridsz{}_bmratio{}_tsteps{}"
        self.aberdir = os.path.join(self.testdir, self.aberroot, aberdir)  # full path to FITS files

        self.prescopyroot = 'prescription'
        prescopydir = "{}"
        self.prescopydir = os.path.join(self.testdir, self.prescopyroot, prescopydir)  # copy of the prescription

        self.device = os.path.join(self.testdir, 'device.pkl')  # detector metadata

    def update_testname(self, new_name='example2'):
        self.__init__(datadir=self.datadir, testname=new_name)

    def update_datadir(self, new_root=f'{str(Path.home())}/MKIDSim/'):
        """Update base directory where  simulation data and output files are saved.
        Remove old directory tree first if empty"""
        self.__init__(datadir=new_root, testname=self.testname)

    def makedir(self):
        #print(self.datadir, self.testdir,  self.scidir)
        if not os.path.isdir(self.datadir):
            os.makedirs(self.datadir, exist_ok=True)
        if not os.path.isdir(self.testdir):
            os.makedirs(self.testdir, exist_ok=True)

    def __iter__(self):
        for attr, value in self.__dict__.items():
            yield attr, value

    def __name__(self):
        return self.__str__().split(' ')[0].split('.')[-1]


class Simulation_params:
    """
    Default parameters for outputs of the simulation. What plots you want to see etc

    """
    def __init__(self):
        self.timing = True  # True will print timing statements in run_medis()
        self.num_processes = 1  # multiprocessing.cpu_count()

        # Grid Sizing/Sampling Params
        self.beam_ratio = 0.5  # parameter dealing with the sampling of the beam in the pupil/focal
                                # plane vs grid size. See Proper manual pgs 36-37 and 71-74 for discussion
        self.grid_size = 512  # creates a nxn array of samples of the wavefront
                              # must be bigger than the beam size to avoid FT effects at edges; must be factor of 2
                              # NOT the size of your detector/# of pixels
        self.maskd_size = 256  # will truncate grid_size to this range (avoids FFT artifacts)
                               # set to grid_size if undesired
        self.focused_sys = False  # use this to turn scaling of beam ratio by wavelength on/off
                        # turn on/off as necessary to ensure optics in focal plane have same sampling at each
                        # wavelength. Can check focal plane sampling in the proper perscription with opx.check_sampling
                        # see Proper manual pg 36 for more info

        # Timing Params
        self.closed_loop = False  # if false (open loop), then initiate multiprocessing for individual timesteps
        self.sample_time = 0.01  # [s] seconds per timestep/frame. used in atmosphere evolution, etc
        self.ao_delay = 0  # [tstep] number of timesteps of delay
        self.startframe = 0  # useful for things like RDI
        self.numframes = 1  # number of timesteps in the simulation
        self.quick_companions = False  # this bool determines if companions are generated by simple shift and scaling
        self.quick_detect = False  # generate mkid spectral cube sequence by spatial scaling and intensity scaling datacube (no photon quantization or arteacts)

        # Plotting Params
        self.show_spectra = False  # Plot spectral cube at single timestep
        self.show_wframe = True  # Plot white light image frame
        self.show_tseries = False  # Plot full timeseries of white light frames
        self.spectra_cols = 2  # number of subplots per row in view_spectra
        self.tseries_cols = 2  # number of subplots per row in view_timeseries

        # Reading/Saving Params
        self.save_to_disk = False  # Saves observation sequence (timestep, wavelength, x, y)
        self.save_list = ['detector']  # list of locations in optics train to save
        self.skip_functions = []  # list of functions not to be applied universally by optics.Wavefronts.loop_over_function
        self.memory_limit = 10  # number of giga-bytes for sixcube of complex fields before chunking happens
        self.checkpointing = 500  # int or None number of timesteps before complex fields sixcube is saved
                                 # minimum of this and max allowed steps for memory reasons takes priority
        self.chunking = False  # chunking is neccessary if the full fields tensor does not fit in memory. RunMedis should set this automatically
        self.auto_load = False  # controls whether RunMedis checks with the user before it loads an old simulaiton it knows to have mismatching parameters
        self.degrade_photons = True

        self.verbose = True
        self.debug = False

    def __iter__(self):
        for attr, value in self.__dict__.items():
            yield attr, value

    def __name__(self):
        return self.__str__().split(' ')[0].split('.')[-1]


class Astro_params:
    """
    Default parameters for the astronomical system under investigation
    exposure_time, startframe and numframes may seem a bit out of place here. Perhaps this class could be renamed
    """
    def __init__(self):
        # Companion Object Params
        self.star_flux = int(1e5)  # A 5 apparent mag star 1e6 cts/cm^2/s
        self.companion = False
        self.contrast = []
        self.C_spec = 1.5  # the gradient of the increase in contrast towards shorter wavelengths
        self.companion_xy = [[-1.0e-6, 1.0e-6]]  # [m] initial location (no rotation)
        self.spectra = [None, None]  # list of floats for Teff of Planck spectrum that scales the wavefronts

        # Wavelength and Wavefront Array Settings
        # In optics_propagate(), proper initially takes N  discreet wavelengths evenly spaced in wvl_range, where N is
        # given by ap.n_wvl_init. Later, in gen_timeseries(), the 3rd axis of the spectral cube is interpolated so that
        # there are ap.n_wvl_final over the range in ap.wvl_range.
        self.n_wvl_init = 3  # initial number of wavelength bins in spectral cube (later sampled by MKID detector)
        self.n_wvl_final = None  # final number of wavelength bins in spectral cube after interpolation (None sets equal to n_wvl_init)
        self.interp_wvl = True  # Set to interpolate wavelengths from ap.n_wvl_init to ap.n_wvl_final
        self.wvl_range = np.array([800, 1500]) / 1e9  # wavelength range in [m]
            # eg. DARKNESS band is [800, 1500], J band =  [1100,1400])

    def __iter__(self):
        for attr, value in self.__dict__.items():
            yield attr, value

    def __name__(self):
        return self.__str__().split(' ')[0].split('.')[-1]


class Telescope_params:
    """
    This contains most of the parameters you will probably modify when running tests
    """
    def __init__(self):
        # Optics + Detector
        self.prescription = 'Subaru_SCExAO'#'general_telescope'#
        self.entrance_d = 5  # 7.971  # [m] telescope enterence pupil diameter in meters
        self.fnum_primary = 12  # f-number of primary
        self.flen_primary = 25  # [m] focal length of primary

        self.use_atmos = False  # Create and apply atmospheric maps (units in phase delay)
        self.obscure = False  # Obscure the primary by the secondary mirror
        self.legs_frac = 0.03  # fractional width of the legs relative to the secondary mirror diameter

        # AO System Settings
        self.use_ao = True  # if False, and DM returns an idealized 'flat'
        self.ao_act = 60  # number of actuators on the DM on one axis (proper only models nxn square DMs)
        self.piston_error = False  # flag for allowing error on DM surface shape
        self.fit_dm = True  # flag to use DM surface fitting (see proper manual pg 52, the FIT switch)
        self.satelite_speck = {'apply': False, 'phase': np.pi / 5., 'amp': 12e-9, 'xloc': 12, 'yloc': 12}

        # Ideal Detector Params (not bothering with MKIDs yet)
        #TODO Rupert should these be moved into mkid params? Are they still used?
        # self.detector = 'ideal'  # 'MKIDs'
        # self.array_size = np.array([129, 129])  # np.array([125,80])
        # self.wavecal_coeffs = [1.e9 / 12, -157]  # assume linear for now 800nm = -90deg, 1500nm = -30deg
                                                # used to make phase cubes. I assume this has something to do with the
                                                # QE of MKIDs at different wavelengths?
        self.pix_shift = [0, 0]  # False?  Shifts the central star to off-axis (mimics conex mirror, tip/tilt error)
        # self.platescale = 13.61  # mas # have to run get_sampling at the focus to find this

        # Aberrations
        self.servo_error = [0, 1]  # [0,1] # False # No delay and rate of 1/frame_time
        self.abertime = 0.5  # time scale of optic aberrations in seconds

        self.lens_params = None  # None at first then gets updated by prescription module

        # Zernike Settings- see pg 192 for details
        self.zernike_orders = [2, 3, 4]  # Order of Zernike Polynomials to include
        self.zernike_vals = np.array([175, -150, 200])*1.0e-9  # value of Zernike order in nm,
                                                               # must be same length as zernike_orders

        # Coronagraph Settings
        self.cg_type = 'Gaussian'
        self.cg_size = 3  # physical size or lambda/D size
        self.cg_size_units = "l/D"  # "m" or "l/D"
        self.lyot_size = 0.75  # units are in fraction of surface blocked
        self.fl_cg_lens = 200 * self.entrance_d  # m

    def __iter__(self):
        for attr, value in self.__dict__.items():
            yield attr, value

    def __name__(self):
        return self.__str__().split(' ')[0].split('.')[-1]


class MKID_params:
    def __init__(self):
        self.bad_pix = False
        # self.interp_sample=True # avoids the quantization error in creating the datacube
        self.QE_map = None
        self.wavecal_coeffs = [1.e9/12, -157]  # assume linear for now 800nm = -90deg, 1500nm = -30deg
        self.phase_uncertainty = False  # True
        self.phase_background = False
        self.QE_var = False
        self.remove_close = False
        self.dark_counts = False
        self.array_size = np.array([129,129])#np.array([125,80])#np.array([125,125])#
        self.resamp = True
        self.quantize_FCs = False
        # self.total_pix = self.array_size[0] * self.array_size[1]
        self.pix_yield = 0.9
        self.hot_pix = 0  # Number of hot pixels
        self.hot_bright = 1e3  # Number of counts/s a hot pixel registers
        self.dark_pix_frac = 0.1  # Number of dark count pixels
        self.dark_bright = 10  # Number of counts/s on average for dark count pixels
        self.threshold_phase = 0#-30 # quite close to 0, basically all photons will be detected.

        self.max_count = 2500.  # cts/s
        self.dead_time = 0.02#10e-6  # s
        self.bin_time = 2e-3 # minimum time to bin counts for stat-based analysis
        # self.frame_time = 0.001#atm_size*atm_spat_rate/(wind_speed*atm_scale) # 0.0004
        self.total_int = 1 #second
        self.frame_int = 1./20
        self.t_frames = int(self.total_int/self.frame_int)
        self.platescale = 10 *1e-3  # [arcsec/pix]

        # for distributions
        self.res_elements = self.array_size[0]
        self.g_mean = 0.95
        self.g_sig = 0.025
        self.g_spec = -1/10.
        self.bg_mean = 0
        self.bg_sig = 30
        self.r_mean = 1
        self.r_sig = 0.05
        self.R_mean = 50
        self.R_sig = 2
        self.R_spec = -1./10

        self.lod = 8  # 8 pixels in these upsampled images = one lambda/d
        self.nlod = 10  # 3 #how many lambda/D do we want to calculate out to

    def __iter__(self):
        for attr, value in self.__dict__.items():
            yield attr, value

    def __name__(self):
        return self.__str__().split(' ')[0].split('.')[-1]


class Device_params:
    """
    This is different from MKID_params in that it contains an instance of these random multidimensional parameters
    Perhaps it could be part of MKID_params
    """
    def __init__(self):
        self.QE_map = None
        self.Rs = None
        self.sigs = None
        self.basesDeg = None
        self.hot_pix = None
        self.dark_pix_frac = None

    def __iter__(self):
        for attr, value in self.__dict__.items():
            yield attr, value

    def __name__(self):
        return self.__str__().split(' ')[0].split('.')[-1]


class Atmos_params():
    """
    default params for making atmospheric models

    original use was to use CAOS, but has now been changed to use hcipy (2019)
    :model: 'single', 'hcipy_standard', 'evolving'
             evolving->apply variation to some parameter

    hcipy still assumes frozen flow as turbulent layers. more here: https://hcipy.readthedocs.io/en/latest/index.html
    """
    def __init__(self):
        self.model = 'single'  # single|hcipy_standard|evolving
        self.cn_sq = 0.22 * 1e-12  # magnitude of perturbations within single atmos layer, at single wavelength
        self.L0 = 10  # outer scale of the model that sets distance of layers (not boundary). used in Kalmogorov model
        self.vel = 5  # velocity of the layer in m/s
        self.h = 100  # scale height in m
        self.fried = 0.2  # m
        self.tau = 0.01 #0.1  # correlation time in seconds of atmopshere
        self.std = 2
        self.correlated_sampling = False

    def __iter__(self):
        for attr, value in self.__dict__.items():
            yield attr, value

    def __name__(self):
        return self.__str__().split(' ')[0].split('.')[-1]


# Creating class structures
ap = Astro_params()
tp = Telescope_params()
sp = Simulation_params()
atmp = Atmos_params()
mp = MKID_params()
dp = Device_params()
iop = IO_params()  # Must call this last since IO_params uses ap and tp

# Turning off messages from Proper
proper.print_it = False
proper.use_cubic_conv = True
# print(proper.__version__)
# proper.prop_init_savestate()




"""
atmosphere.py

For mini-medis, code copied over from atmos.py and aberrations.py from Rupert's original MEDIS

Contains functions that will create, load, save, and apply atmospheric maps
"""
import numpy as np
import os
import astropy.io.fits as fits
import hcipy
import proper
from skimage.restoration import unwrap_phase
# from mkidpipeline.speckle.genphotonlist_IcIsIr import corrsequence
from scipy import special

from medis.params import iop, ap, tp, sp, atmp
from medis.utils import dprint, clipped_zoom
from medis.optics import circular_mask

def recursion(r,g, f, sqrt1mf2, n):
    for i in range(1, n):
        r[i] = r[i - 1]*f + g[i]*sqrt1mf2
    return r

def corrsequence(Ttot, tau):
    """
    Generate a sequence of correlated Gaussian noise, correlation time
    tau.  Algorithm is recursive and from Markus Deserno.  The
    recursion is implemented as an explicit for loop but has a lower
    computational cost than converting uniform random variables to
    modified-Rician random variables.

    Arguments:
    Ttot: int, the total integration time in microseconds.
    tau: float, the correlation time in microseconds.

    Returns:
    t: a list of integers np.arange(0, Ttot)
    r: a correlated Gaussian random variable, zero mean and unit variance, array of length Ttot

    """

    t = np.arange(Ttot)
    g = np.random.normal(0, 1, Ttot)
    r = np.zeros(g.shape)
    f = np.exp(-1. / tau)
    sqrt1mf2 = np.sqrt(1 - f ** 2)
    r = recursion(r, g, f, sqrt1mf2, g.shape[0])

    return t, r

def gen_atmos(plot=False, debug=True):
    """
    generates atmospheric phase distortions using hcipy (updated from original using CAOS)

    read more on HCIpy here: https://hcipy.readthedocs.io/en/latest/index.html
    In hcipy, the atmosphere evolves as a function of time, specified by the user. User can thus specify the
    timescale of evolution through both velocity of layer and time per step in the obs_sequence, in loop for
    medis_main.gen_timeseries().

    :param plot: turn plotting on or off
    :return:

    todo add simple mp.Pool.map code to make maps in parrallel
    """
    if tp.use_atmos is False:
        pass  # only make new atmosphere map if using the atmosphere
    else:

        if sp.verbose: dprint("Making New Atmosphere Model")
        # Saving Parameters
        # np.savetxt(iop.atmosconfig, ['Grid Size', 'Wvl Range', 'Number of Frames', 'Layer Strength', 'Outer Scale', 'Velocity', 'Scale Height', cp.model])
        # np.savetxt(iop.atmosconfig, ['ap.grid_size', 'ap.wvl_range', 'ap.numframes', 'atmp.cn_sq', 'atmp.L0', 'atmp.vel', 'atmp.h', 'cp.model'])
        # np.savetxt(iop.atmosconfig, [ap.grid_size, ap.wvl_range, ap.numframes, atmp.cn_sq, atmp.L0, atmp.vel, atmp.h, cp.model], fmt='%s')

        wsamples = np.linspace(ap.wvl_range[0], ap.wvl_range[1], ap.n_wvl_init)
        wavefronts = []

        ##################################
        # Initiate HCIpy Atmosphere Type
        ##################################
        pupil_grid = hcipy.make_pupil_grid(sp.grid_size, tp.entrance_d)
        if atmp.model == 'single':
            layers = [hcipy.InfiniteAtmosphericLayer(pupil_grid, atmp.cn_sq, atmp.L0, atmp.vel, atmp.h, 2)]
        elif atmp.model == 'hcipy_standard':
            # Make multi-layer atmosphere
            # layers = hcipy.make_standard_atmospheric_layers(pupil_grid, atmp.L0)
            heights = np.array([500, 1000, 2000, 4000, 8000, 16000])
            velocities = np.array([10, 10, 10, 10, 10, 10])
            Cn_squared = np.array([0.2283, 0.0883, 0.0666, 0.1458, 0.3350, 0.1350]) * 3.5e-12

            layers = []
            for h, v, cn in zip(heights, velocities, Cn_squared):
                layers.append(hcipy.InfiniteAtmosphericLayer(pupil_grid, cn, atmp.L0, v, h, 2))

        elif atmp.model == 'evolving':
            raise NotImplementedError
        atmos = hcipy.MultiLayerAtmosphere(layers, scintilation=False)

        for wavelength in wsamples:
            wavefronts.append(hcipy.Wavefront(hcipy.Field(np.ones(pupil_grid.size), pupil_grid), wavelength))

        if atmp.correlated_sampling:
            # Damage Detection and Localization from Dense Network of Strain Sensors

            # fancy sampling goes here
            normal = corrsequence(sp.numframes, atmp.tau/sp.sample_time)[1] * atmp.std
            uniform = (special.erf(normal / np.sqrt(2)) + 1)

            times = np.cumsum(uniform) * sp.sample_time

            if debug:
                import matplotlib.pylab as plt
                plt.plot(normal)
                plt.figure()
                plt.plot(uniform)
                plt.figure()
                plt.hist(uniform)
                plt.figure()
                plt.plot(np.arange(0, sp.numframes * sp.sample_time, sp.sample_time))
                plt.plot(times)
                plt.show()
        else:
            times = np.arange(0, sp.numframes * sp.sample_time, sp.sample_time)

        ###########################################
        # Evolving Wavefront using HCIpy tools
        ###########################################
        for it, t in enumerate(times):
            atmos.evolve_until(t)
            for iw, wf in enumerate(wavefronts):
                wf2 = atmos.forward(wf)

                filename = get_filename(it, wsamples[iw], (iop.atmosdir, sp.sample_time,
                                                           atmp.model))
                if sp.verbose: dprint(f"atmos file = {filename}")
                hdu = fits.ImageHDU(wf2.phase.reshape(sp.grid_size, sp.grid_size))
                hdu.header['PIXSIZE'] = tp.entrance_d/sp.grid_size
                hdu.writeto(filename, overwrite=True)

                if plot and iw == 0:
                    import matplotlib.pyplot as plt
                    from medis.twilight_colormaps import sunlight
                    plt.figure()
                    plt.title(f"Atmosphere Phase Map t={t} lambda={eformat(wsamples[iw], 3, 2)}")
                    hcipy.imshow_field(wf2.phase, cmap=sunlight)
                    plt.colorbar()
                    plt.show(block=True)


def add_atmos(wf, it, param_tup=None, spatial_zoom=False):
    """
    creates a phase offset matrix for each wavelength at each time step,
    sampled from the atmosphere generated by hcipy

    HCIpy generates an atmosphere with given parameters. The returned field is in units of phase delay for each
    wavelength. prop_add_phase wants the phase delay to be in units of meters for each wavelength, so we convert to
    meters via the wavelength/np.pi

    :param wf: a single (2D) wfo.wf_collection[iw,ib] at one wavelength and object
    :param it: timestep# in obs_sequence. Comes from medis_main.gen_timeseries()
    :return: nothing returned, wfo is modified with proper.prop_add_phase
    """
    if tp.use_atmos is False:
        pass  # don't do anything. Putting this type of check here allows universal toggling on/off rather than
                # commenting/uncommenting in the proper perscription
    else:
        wavelength = wf.lamda  # the .lamda comes from proper, not from Wavefronts class

        # Check for Existing File)
        atmos_map = get_filename(it, wavelength, param_tup)
        if not os.path.exists(atmos_map):
            # todo remove when all test scripts use the new format
            # TODO check for new file name convention in addition to directory name
            print('atmospheres should be created at the beginng, not on the fly')
            raise NotImplementedError

        atm_map = fits.open(atmos_map)[1].data
        atm_map = unwrap_phase(atm_map)
        atm_map *= wavelength/(2*np.pi)  # converts atmosphere in units of phase delay (rad) into distance (m)

        if spatial_zoom:
            scale = ap.wvl_range[0] / wavelength
            atm_map = clipped_zoom(atm_map, scale)
            h, w = atm_map.shape[:2]
            mask = circular_mask(h, w, radius=scale * h * sp.beam_ratio / 2)
            atm_map[~mask] = 0
            atm_map[mask] -= np.mean(atm_map[mask])  # remove bias from spatial stretching

        proper.prop_add_phase(wf, atm_map)


def rotate_atmos(wf, it):
    time = it * ap.sample_time
    rotate_angle = tp.rot_rate * time
    wf.wfarr = proper.prop_shift_center(wf.wfarr)
    wf.wfarr = proper.prop_rotate(wf.wfarr, rotate_angle)
    wf.wfarr = proper.prop_shift_center(wf.wfarr)


def eformat(wvl, prec, exp_digits):
    """
    reformats wavelength into scientific notation in meters

    :param wvl: wavelength float
    :param prec: precision
    :param exp_digits: number of digits in the exponent
    :return:
    """
    s = "%.*e" % (prec, wvl)
    mantissa, exp = s.split('e')
    # add 1 to digits as 1 is taken by sign +/-
    return "%se%+0*d" % (mantissa, exp_digits + 1, int(exp))


def get_filename(it, lamda, param_tup=None):
    """
    returns the atmosphere map names in the format location/atmos_t<time>_<model>_wvl<wavelength>
    example output:

    :param it: time index
    :param lamda: wavelength, comes from metadata in proper wavefront object (ie wfo.wf_collection.lamda)
    :return:
    """
    wave = eformat(lamda, 3, 2)
    if param_tup:
        atmosdir, sample_time, model = param_tup
    else:
        atmosdir, sample_time, model = iop.atmosdir, sp.sample_time, atmp.model

    sigfig = int(np.abs(np.floor(np.log10(sp.sample_time))))

    return f'{atmosdir}/atmos_t{sample_time*it:.6f}_{model}_wvl{wave}.fits'


"""
aberrations.py
mostly copied over from the original by Rupert in MEDIS

collection of functions that deal with making aberration maps in proper for a given optical element.

Two initialize functions set up the ques for creating, saving, and reading FITs files of aberration maps. In general,
for an optical element in the telescope optical train, a aberration map is generated in Proper using prop_psd_errormap.
The map is saved as a FITs file and used for every wavelength and timestep in the observation sequence. Ideally this
will be updated for quasi-static aberrations, where the aberrations evolve over some user-defined timescale.
"""
import numpy as np
import proper
import os
import pickle

from medis.params import iop
from medis.utils import *


################################################################################################################
# Aberrations
################################################################################################################
def generate_maps(aber_vals, lens_diam, lens_name='lens', quasi_static=False):
    """
    generate PSD-defined aberration maps for a lens(mirror) using Proper

    Use Proper to generate an 2D aberration pattern across an optical element. The amplitude of the error per spatial
     frequency (cycles/m) across the surface is taken from a power spectral density (PSD) of statistical likelihoods
     for 'real' aberrations of physical optics.
    parameters defining the PSD function are specified in tp.aber_vals. These limit the range for the constants of the
     governing equation given by PSD = a/ [1+(k/b)^c]. This formula assumes the Terrestrial Planet Finder PSD, which is
     set to TRUE unless manually overridden line-by-line. As stated in the proper manual, this PSD function general
      under-predicts lower order aberrations, and thus Zernike polynomials can be added to get even more realistic
      surface maps.
    more information on a PSD error map can be found in the Proper manual on pgs 55-60

    Note: Functionality related to OOPP (out of pupil plane) optics has been removed. There is only one surface
    simulated for each optical surface

    :param aber_vals: dictionary? of values to use in the equation that generates the aberration map. The dictionary
        should contain 3 entries that get sent to proper.prop_psd_errormap. More information can be found on Proper
        Manual pg 56
    :param lens_diam: diameter of the lens/mirror to generate an aberration map for
    :param lens_name: name of the lens, for file naming
    :return: will create a FITs file in the folder specified by iop.quasi for each optic (and  timestep in the case
     of quasi-static aberrations)
    """
    # TODO add different timescale aberations
    if sp.verbose: dprint(f'Generating optic aberration maps using Proper at directory {iop.aberdir}')
    if not os.path.isdir(iop.aberdir):
        # todo remove when all test scripts use the new format
        raise NotImplementedError
        print('aberration maps should be created at the beginng, not on the fly')
        os.makedirs(iop.aberdir, exist_ok=True)

    # create blank lens wavefront for proper to add phase to
    wfo = proper.prop_begin(lens_diam, 1., sp.grid_size, sp.beam_ratio)
    aber_cube = np.zeros((1, sp.grid_size, sp.grid_size))

    # Randomly select a value from the range of values for each constant
    # rms_error = np.random.normal(aber_vals['a'][0], aber_vals['a'][1])
    # c_freq = np.random.normal(aber_vals['b'][0], aber_vals['b'][1])  # correlation frequency (cycles/meter)
    # high_power = np.random.normal(aber_vals['c'][0], aber_vals['c'][1])  # high frequency falloff (r^-high_power)

    rms_error, c_freq, high_power = aber_vals

    # perms = np.random.rand(sp.numframes, sp.grid_size, sp.grid_size)-0.5
    # perms *= 1e-7

    phase = 2 * np.pi * np.random.uniform(size=(sp.grid_size, sp.grid_size)) - np.pi
    aber_cube[0] = proper.prop_psd_errormap(wfo, rms_error, c_freq, high_power, TPF=True, PHASE_HISTORY=phase)
        # PHASE_HISTORY stuff is a kwarg Rupert added to a proper.prop_pds_errormap in proper_mod that helps
        #  ennable the small perturbations to the phase aberrations over time (quasi-static aberration evolution)
        #  however, this may not be implemented here, and the functionality may not be robust. It has yet to be
        #  verified in a robust manner. However, I am not sure it is being used....? KD 10-15-19
    # TODO verify this and add qusi-static functionality

    filename = f"{iop.aberdir}/t{0}_{lens_name}.fits"
    #dprint(f"filename = {filename}")
    if not os.path.isfile(filename):
        saveFITS(aber_cube[0], filename)

    if quasi_static:
        # todo implement monkey patch of proper.prop_psd_error that return phase too so it can be incremented with correlated gaussian noise
        # See commit 48c77c0babd5c6fcbc681b4fdd0d2db9cf540695 for original implementation of this code
        raise NotImplementedError


def add_aber(wf, aberdir=None, step=0, lens_name=None):
    """
    loads a phase error map and adds aberrations using proper.prop_add_phase
    if no aberration file exists, creates one for specific lens using generate_maps

    :param wf: a single (2D) wfo.wf_collection[iw,io] at one wavelength and object
    :param d_lens: diameter (in m) of lens (only used when generating new aberrations maps)
    :param step: is the step number for quasistatic aberrations
    :param lens_name: name of the lens, used to save/read in FITS file of aberration map
    :return returns nothing but will act upon a given wavefront and apply new or loaded-in aberration map
    """
    # TODO this does not currently loop over time, so it is not using quasi-static abberations.
    if tp.use_aber is False:
        pass  # don't do anything. Putting this type of check here allows universal toggling on/off rather than
              # commenting/uncommenting in the proper perscription
    else:
    # dprint("Adding Abberations")

        # Load in or Generate Aberration Map
        # iop.aberdata = f"gridsz{sp.grid_size}_bmratio{sp.beam_ratio}_tsteps{sp.numframes}"
        # iop.aberdir = os.path.join(iop.testdir, iop.aberroot, iop.aberdata)
        filename = f"{iop.aberdir}/t{0}_{lens_name}.fits"
        # print(f'Adding Abberations from {filename}')

        # if not os.path.isfile(filename):
        #     generate_maps(aber_vals, d_lens, lens_name)

        phase_map = readFITS(filename)
        proper.prop_add_phase(wf, phase_map)     # Add Phase Map


def add_zern_ab(wf, zern_order=[2,3,4], zern_vals=np.array([175,-150,200])*1.0e-9):
    """
    adds low-order aberrations from Zernike polynomials

    see Proper Manual pg 192 for full details
    see good example in Proper manual pg 51
    quote: These [polynomials] form an orthogonal set of aberrations that include:
     wavefront tilt, defocus, coma, astigmatism, spherical aberration, and others

    Orders are:
    1 Piston
    2 X tilt
    3 Y tilt
    4 Focus
    5 45º astigmatism
    6 0º astigmatism
    7 Y coma
    8 X coma
    9 Y clover (trefoil)
    10 X clover (trefoil)
    """
    if not tp.add_zern:
        pass
    else:
        proper.prop_zernikes(wf, zern_order, zern_vals)


def randomize_zern_values(zern_orders):
    """
    selects a value at random for the selected zernike orders

    The ranges specified are all in units of m. From Proper manual pg 192-193
    'The values must be in meters of RMS phase error or dimensionless RMS amplitude error.'

    :param zern_orders: array containing integers of zernike orders to apply
    :return: array with randomly selected values in a pre-determined range, per zernike order specified in zern_orders
    """
    # Set Range
    z_range = {
        'r1': [30e-9, 60e-9],  # m of RMS amplitude error
        'r2': [30e-9, 40e-9],  # X tilt
        'r3': [30e-9, 40e-9],  # Y tilt
        'r4': [60e-9, 120e-9],  # focus
        'r5': [30e-9, 1200e-9],
        'r6': [30e-9, 1200e-9],
        'r7': [30e-9, 1200e-9],
        'r8': [30e-9, 1200e-9],
        'r9': [30e-9, 1200e-9],
        'r10': [30e-9, 1200e-9],
    }

    zern_vals = np.zeros(len(zern_orders))
    for iw, w in enumerate(zern_orders):
        if w == 1:
            zern_vals[iw] = np.random.normal(z_range['r1'][0], z_range['r1'][1])
        elif w == 2:
            zern_vals[iw] = np.random.normal(z_range['r2'][0], z_range['r2'][1])
        elif w == 3:
            zern_vals[iw] = np.random.normal(z_range['r3'][0], z_range['r3'][1])
        elif w == 4:
            zern_vals[iw] = np.random.normal(z_range['r4'][0], z_range['r4'][1])
        elif w == 5:
            zern_vals[iw] = np.random.normal(z_range['r5'][0], z_range['r5'][1])
        elif w == 6:
            zern_vals[iw] = np.random.normal(z_range['r6'][0], z_range['r6'][1])
        elif w == 7:
            zern_vals[iw] = np.random.normal(z_range['r7'][0], z_range['r7'][1])
        elif w == 8:
            zern_vals[iw] = np.random.normal(z_range['r8'][0], z_range['r8'][1])
    return zern_vals



###############################################################################################
# Depricated--but potenetially useful for quasi-static aberration implementation later
###############################################################################################
# def initialize_CPA_meas():
#     required_servo = int(tp.servo_error[0])
#     required_band = int(tp.servo_error[1])
#     required_nframes = required_servo + required_band + 1
#     CPA_maps = np.zeros((required_nframes, ap.n_wvl_init, sp.grid_size, sp.grid_size))
#
#     with open(iop.CPA_meas, 'wb') as handle:
#         pickle.dump((CPA_maps, np.arange(0, -required_nframes, -1)), handle, protocol=pickle.HIGHEST_PROTOCOL)
#
#
# def initialize_NCPA_meas():
#     Imaps = np.zeros((4, sp.grid_size, sp.grid_size))
#     phase_map = np.zeros((tp.ao_act, tp.ao_act))  # np.zeros((sp.grid_size,sp.grid_size))
#     with open(iop.NCPA_meas, 'wb') as handle:
#         pickle.dump((Imaps, phase_map, 0), handle, protocol=pickle.HIGHEST_PROTOCOL)





"""
todo h5 won't let you save chunks>4GB. Add that check to max_chunk function (in addition to one users specific memory check)

"""

import os
import sys
import numpy as np
import importlib
import multiprocessing
import time
import glob
import pickle
import shutil
import tables

import proper
import medis.atmosphere as atmos
from medis.plot_tools import grid
import medis.CDI as cdi
import medis.utils as mu
import medis.optics as opx
import medis.aberrations as aber
from medis.params import sp, ap, tp, iop
from medis.CDI import cdi


class Telescope:
    """
    Creates a simulation for the telescope to create a series of complex electric fields

    During initialisation a backup of the pthon PROPER prescription is copied to the testdir, atmisphere maps and
    aberration maps are created, serialisation and memory requirements are tested, and the cdi vairable initialised

    Resulting file structure:
    datadir
        testdir
            params.pkl                     <--- input
            prescription                   <--- new
                {prescriptionname}         <--- new
                    {prescriptionname}.py  <--- new
            aberrations                    <--- new
                {aberationparams}          <--- new
                    {lensname}0.fits       <--- new
                    ...                    <--- new
            atmosphere                     <--- new
                {atmosphereparams}         <--- new
                    {atmos}0.fits          <--- new
                    ...                    <--- new
            fields.h5                      <--- output


    input
    params dict
        collection of the objects in params.py

    :return:
    self.cpx_sequence ndarray
        complex tensor of dimensions (n_timesteps, n_saved_planes, n_wavelengths, n_stars/planets, grid_size, grid_size)
    """

    def __init__(self, usesave=True):
        self.usesave=usesave

        self.fields_exists = True if os.path.exists(iop.fields) else False  # just the fields ndarray
        self.save_exists = True if os.path.exists(iop.telescope) else False  # the whole telescope object

        # if self.fields_exists:
        #     self.load_fields()
        if self.save_exists:
            print(f"\nLoading telescope instance from\n\t{iop.telescope}\n")
            with open(iop.telescope, 'rb') as handle:
                load = pickle.load(handle)
                self.__dict__ = load.__dict__  # replace the contents
                self.save_exists = True
        else:
            # copy over the prescription
            iop.prescopydir = iop.prescopydir.format(tp.prescription)
            self.target = os.path.join(iop.prescopydir, tp.prescription+'.py')

            prescriptions = iop.prescriptions_root
            fullprescription = glob.glob(os.path.join(prescriptions, '**', tp.prescription+'.py'),
                                         recursive=True)
            if len(fullprescription) == 0:
                raise FileNotFoundError
            elif len(fullprescription) > 1:
                print(f'Multiple precriptions at {fullprescription}')
                raise FileExistsError

            fullprescription = fullprescription[0]
            if os.path.exists(self.target):
                print(f'Using prescription {fullprescription}\n')
            else:
                print(f"Copying over prescription {fullprescription}\n")

                if not os.path.isdir(iop.prescopydir):
                    os.makedirs(iop.prescopydir, exist_ok=True)
                shutil.copyfile(fullprescription, self.target)

            sys.path.insert(0, os.path.dirname(fullprescription))  # load from the original prescription incase user is editting
            pres_module = importlib.import_module(tp.prescription)
            tp.__dict__.update(pres_module.tp.__dict__)  #  update tp with the contents of the prescription

            # initialize atmosphere
            iop.atmosdir = iop.atmosdir.format(sp.grid_size, sp.beam_ratio, sp.numframes)
            if glob.glob(iop.atmosdir+'/*.fits') and sp.verbose:
                print(f"Atmosphere maps already exist at \n\t{iop.atmosdir}"
                      f" \n... skipping generation\n\n")
            else:
                if not os.path.isdir(iop.atmosdir):
                    os.makedirs(iop.atmosdir, exist_ok=True)
                atmos.gen_atmos()

            # initialize aberrations
            iop.aberdir = iop.aberdir.format(sp.grid_size, sp.beam_ratio, sp.numframes)
            if glob.glob(iop.aberdir + '/*.fits') and sp.verbose:
                print(f"Aberration maps already exist at \n\t{iop.aberdir} "
                      f"\n... skipping generation\n\n")
            else:
                if not os.path.isdir(iop.aberdir):
                    os.makedirs(iop.aberdir, exist_ok=True)
                for lens in tp.lens_params:
                    aber.generate_maps(lens['aber_vals'], lens['diam'], lens['name'])

            # check if can do parrallel
            if sp.closed_loop or sp.ao_delay:
                print(f"closed loop or ao delay means sim can't be parrallelized in time domain. Forcing serial mode")
                self.parrallel = False
            else:
                self.parrallel = sp.num_processes > 1

            # ensure contrast is set properly
            if sp.quick_companions:
                # purpose of this code to add option to shift and scale an unocculed star and use that for several sources vastly decreasing compute time
                raise NotImplementedError
                ap.contrast = range(2)  # give it length two since all the planets will be collapsed into one slice

            if ap.companion is False:
                ap.contrast = []

            if ap.companion and len(ap.spectra) != len(ap.contrast)+1 != len(ap.companion_xy)+1:
                print(f'Please ensure number of sources is consistent: {len(ap.spectra)}, {len(ap.contrast)+1}, {len(ap.companion_xy)+1}')
                raise IndexError

            if not isinstance(ap.n_wvl_final, int):
                ap.n_wvl_final = ap.n_wvl_init

            # determine if can/should do all in memory
            max_steps = self.max_chunk()
            self.fieldssize = self.timestep_size * sp.numframes
            if sp.verbose: print(f"File total size should be {self.fieldssize * 1e-6} MB")
            checkpoint_steps = max_steps if sp.checkpointing is None else sp.checkpointing
            self.chunk_steps = int(min([max_steps, sp.numframes, checkpoint_steps]))
            if sp.verbose: print(f'Using time chunks of size {self.chunk_steps}')
            self.num_chunks = sp.numframes / self.chunk_steps

            if self.num_chunks > 1:
                print('Simulated data too large for dynamic memory. Storing to disk as the sim runs')
                sp.chunking = True

            self.markov = sp.chunking or self.parrallel  # independent timesteps
            nonmarkov = sp.ao_delay or sp.closed_loop  # dependent timesteps
            # if both true
            assert self.markov + nonmarkov != 2, "Confliciting modes. Request requires the timesteps be both dependent and independent"

            # neither true
            if self.markov + nonmarkov == 0:
                print("No mode specfified defaulting to markov (time independent)")
                self.markov = True

            modes = [sp.chunking, sp.ao_delay, self.parrallel,
                     sp.closed_loop]

            # Initialize List of phases of CDI probes to apply
            cdi.gen_phaseseries()
            if cdi.use_cdi:
                if hasattr(tp, 'act_tweeter'):
                    cdi.init_cout(tp.act_tweeter)
                else:
                    cdi.init_cout(tp.ao_act)

            # Remove AO planes from save_list if use_ao is False
            if not tp.use_ao and 'woofer' in sp.save_list:
                sp.save_list.remove('woofer')
            if not tp.use_ao and 'tweeter' in sp.save_list:
                sp.save_list.remove('tweeter')

    def __call__(self, *args, **kwargs):
        """ Take the observation (generate the fields sequence) """
        if self.fields_exists:
            self.load_fields()
        else:
            print('\n\n\tBeginning Telescope Simulation with MEDIS\n\n')
            start = time.time()

            self.create_fields()

            print('\n\n\tMEDIS Telescope Run Completed\n')
            finish = time.time()
            print(f'Time elapsed: {(finish - start) / 60:.2f} minutes')

            self.pretty_sequence_shape()
            self.save_exists = True

            if self.usesave and self.fieldssize > 4e9:  # bug with pickling on Mac preventing dump beyond 4GB
                print(f"\nSaving telescope instance at\n\n\t{iop.telescope}\n")
                with open(iop.telescope, 'wb') as handle:
                    pickle.dump(self, handle, protocol=pickle.HIGHEST_PROTOCOL)

        if len(self.cpx_sequence) < sp.numframes:
            print('Returning the final time chunk instead of the full duration. Increase chunk size if you want full '
                  'duration')
        dataproduct = {'fields': np.array(self.cpx_sequence), 'sampling': self.sampling}
        return dataproduct

    def max_chunk(self):
        """
        Determines the maximum duration each chunk can be to fit within the memory limit

        :return: integer
        """
        self.timestep_size = len(sp.save_list) * ap.n_wvl_final * \
                        (1 + len(ap.contrast)) * sp.grid_size**2 * 32 / 8  # in Bytes

        max_chunk = sp.memory_limit*1e9 // self.timestep_size
        print(f'Each timestep is predicted to be {self.timestep_size/1.e6} MB, meaning no more than {max_chunk} time '
              f'steps can fit in the memory at one time')

        return max_chunk

    def create_fields(self):
        """ Create fields tensor for an initialised Telecope instance """

        t0 = sp.startframe
        self.kwargs = {}
        self.cpx_sequence = None

        if self.markov:  # time steps are independent
            ceil_num_chunks = int(np.ceil(self.num_chunks))
            if ceil_num_chunks > 1:
                print('Only partial observation will be in memory at one time')
            final_chunk_size = sp.numframes-int(np.floor(self.num_chunks))*self.chunk_steps
            for ichunk in range(ceil_num_chunks):
                fractional_step = final_chunk_size != 0 and ichunk == ceil_num_chunks-1
                chunk_steps = final_chunk_size if fractional_step else self.chunk_steps

                cpx_sequence = np.empty((chunk_steps, len(sp.save_list),
                                        ap.n_wvl_init, 1 + len(ap.contrast),
                                        sp.grid_size, sp.grid_size),
                                        dtype=np.complex64)
                chunk_range = ichunk * self.chunk_steps + t0 + np.arange(chunk_steps)
                if sp.num_processes == 1:
                    seq_samp_list = [self.run_timestep(t) for t in chunk_range]
                else:
                    print(f'Using multiprocessing of timesteps {chunk_range}')
                    # it appears as though the with statement is neccesssary when recreating Pools like this
                    with multiprocessing.Pool(processes=sp.num_processes) as p:
                        seq_samp_list = p.map(self.run_timestep, chunk_range)
                self.cpx_sequence = np.array([tup[0] for tup in seq_samp_list])
                self.sampling = seq_samp_list[0][1]

                if ap.n_wvl_init < ap.n_wvl_final:
                    self.cpx_sequence = opx.interp_wavelength(self.cpx_sequence, ax=2)
                    self.sampling = opx.interp_sampling(self.sampling)

                if sp.save_to_disk: self.save_fields(self.cpx_sequence)

        else:
            self.cpx_sequence = np.zeros((sp.numframes, len(sp.save_list),
                                          ap.n_wvl_init, 1 + len(ap.contrast),
                                          sp.grid_size, sp.grid_size), dtype=np.complex)
            self.sampling = np.zeros((len(sp.save_list), ap.n_wvl_init))

            for it, t in enumerate(range(t0, sp.numframes + t0)):
                WFS_ind = ['wfs' in plane for plane in sp.save_list]
                if t > sp.ao_delay:
                    self.kwargs['WFS_field'] = self.cpx_sequence[it - sp.ao_delay, WFS_ind, :, 0]
                    self.kwargs['AO_field'] =  self.cpx_sequence[it - sp.ao_delay, AO_ind, :, 0]
                else:
                    self.kwargs['WFS_field'] = np.zeros((ap.n_wvl_init, sp.grid_size,
                                                    sp.grid_size), dtype=np.complex)
                    self.kwargs['AO_field'] = np.zeros((ap.n_wvl_init, sp.grid_size,
                                                        sp.grid_size), dtype=np.complex)
                self.cpx_sequence[it], self.sampling = self.run_timestep(t)

            print('************************')
            if sp.save_to_disk: self.save_fields(self.cpx_sequence)

        # return {'fields': np.array(self.cpx_sequence), 'sampling': self.sampling}

    def run_timestep(self, t):
        self.kwargs['iter'] = t
        return proper.prop_run(tp.prescription, 1, sp.grid_size, PASSVALUE=self.kwargs, QUIET=True)
        # return np.zeros((1, len(sp.save_list), ap.n_wvl_init, 1 + len(ap.contrast),
        #                  sp.grid_size, sp.grid_size), dtype=np.complex64), 0

    def pretty_sequence_shape(self):

        """
        displays data format easier

        :param cpx_sequence: the 6D complex sequence generated by run_medis.telescope
        :return: nicely parsed string of 6D shape--human readable output
        """

        if len(np.shape(self.cpx_sequence)) == 6:
            samps = ['timesteps', 'save planes', 'wavelengths', 'astronomical bodies', 'x', 'y']
            delim = ', '
            print(f"Shape of cpx_sequence = " \
                f"{delim.join([samp + ':' + str(length) for samp, length in zip(samps, np.shape(self.cpx_sequence))])}")
        else:
            print(f'Warning cpx_sequence is not 6D as intented by this function. Shape of tensor ='
                  f' {self.cpx_sequence.shape}')

    def save_fields(self, fields):
        """
        Option to save fields separately from the class pickle save since fields can be huge if the user requests

        :param fields:
            dtype ndarray of complex or float
            fields can be any shape but the h5 dataset can only extended along axis 0
        :return:

        todo convert to pytables for pipeline conda integration
        """

        fields = np.array(fields)
        print(f'Saving fields dims with dimensions {fields.shape} and type {fields.dtype}')
        h5file = tables.open_file(iop.fields, mode="a", title="MEDIS Electric Fields File")
        if "/data" not in h5file:
            ds = h5file.create_earray(h5file.root, 'data', obj=fields)
        else:
            ds = h5file.root.data
            ds.append(fields)

        h5file.close()

    def load_fields(self, span=(0,-1)):
        """ load fields h5

         warning sampling is not currently stored in h5. It is stored in telescope.pkl however
         """
        print(f"Loading fields from {iop.fields}")
        h5file = tables.open_file(iop.fields, mode="r", title="MEDIS Electric Fields File")
        if span[1] == -1:
            self.cpx_sequence = h5file.root.data[span[0]:]
        else:
            self.cpx_sequence = h5file.root.data[span[0]:span[1]]
        h5file.close()
        self.pretty_sequence_shape()
        self.sampling = None

        return {'fields': self.cpx_sequence, 'sampling': self.sampling}


if __name__ == '__main__':
    iop.update_testname('telescope_module_test')

    telescope_sim = Telescope()
    observation = telescope_sim()
    print(observation.keys())
    grid(observation['fields'], logZ=True, nstd=5)



"""
The Camera class simulates an obsservation with an MKID camera given a "fields" tensor (created with Telescope) and
the MKID_params that define the device
"""

import os, sys
import numpy as np
from matplotlib import pyplot as plt
from scipy import interpolate
import pickle
import random
import time

from medis.distribution import *
from medis.params import mp, ap, iop, sp, tp
from medis.utils import dprint
from medis.plot_tools import view_spectra
from medis.telescope import Telescope
from medis.plot_tools import grid, quick2D


class Camera():
    def __init__(self, usesave=False, product='photons'):
        """
        Creates a simulation for the MKID Camera to create a series of photons from the fields sequence

        Resulting file structure:
        datadir
            testdir
                params.pkl         <--- input
                fields.h5          <--- input
                camera             <--- new
                    devicesave.pkl <--- new
                photonlist.h5      <--- output


        input
        fields ndarray complex
            complex tensor of dimensions (n_timesteps, n_saved_planes, n_wavelengths, n_stars/planets, grid_size, grid_size)
        usesave: bool
        product: 'photons' | 'rebinned_cube'

        :return:
        either photon list or rebinned cube

        Functions
        __init__():        loads instance of class if it exists, otherwise loads photonlist if it exist, otherwise gets
                           fields and creates device
        create_device():   create the camera's random device parameters eg locations of dead pix, assigning R to pixels etc
        __call__():        if photons can't be loaded detect photons (combine fields and device params to produce a photon
                           list) and save
        save_photonlist(): save the photonlist in the MKIDPipeline format
        save_instance():   save the whole instance of this class including photons and self.name etc

        """
        assert product in ['photons', 'rebinned_cube'], f"Requested data product {self.product} not supported"

        self.name = iop.camera  # used for saving an loading the instance
        self.usesave = usesave
        self.product = product
        self.rebinned_cube = None
        self.is_wave_cal = False

        self.save_exists = True if os.path.exists(self.name) else False  # whole Camera instance
        self.photontable_exists = True if os.path.exists(iop.photonlist) else False  # just the photonlist

        if self.save_exists and self.usesave:
            load = self.load_instance()
            self.__dict__ = load.__dict__
            self.save_exists = True  # just in case the saved obj didn't have this set to True

        else:
            # create device
            self.create_device()

    def create_device(self):
        """
        Create the camera's random device parameters eg locations of dead pix, assigning R to pixels etc

        :return:
        """
        self.platescale = mp.platescale
        self.array_size = mp.array_size
        self.dark_pix_frac = mp.dark_pix_frac
        self.hot_pix = mp.hot_pix
        self.lod = mp.lod
        self.max_count = mp.max_count
        self.numframes = sp.numframes  # sometimes numframes is different for different cams

        self.QE_map_all = self.array_QE(plot=False)
        # self.max_count = mp.max_count
        self.responsivity_error_map = self.responvisity_scaling_map(plot=False)

        if mp.bad_pix:
            self.QE_map = self.create_bad_pix(self.QE_map_all, plot=False) if mp.pix_yield != 1 else self.QE_map_all

            if mp.dark_counts:
                self.total_dark = sp.sample_time * mp.dark_bright * self.array_size[0] * self.array_size[
                    1] * self.dark_pix_frac * sp.numframes
                self.dark_locs = self.create_false_pix(amount=int(
                    mp.dark_pix_frac * self.array_size[0] * self.array_size[1]))
            if mp.hot_pix:
                self.total_hot = sp.sample_time * mp.hot_bright * self.hot_pix * sp.numframes
                self.hot_locs = self.create_false_pix(amount=mp.hot_pix)

        self.Rs = self.assign_spectral_res(plot=False)
        self.sigs = self.get_R_hyper(self.Rs, plot=False)

        # get_phase_distortions(plot=True)
        if mp.phase_background:
            self.basesDeg = self.assign_phase_background(plot=False)
        else:
            self.basesDeg = np.zeros((self.array_size))

        print('\nInitialized MKID device parameters\n')

    def __call__old(self, fields=None, abs_step=0, finalise_photontable=True, *args, **kwargs):
        if self.photontable_exists and self.usesave:
            self.load_photontable()

        else:
            assert fields is not None, 'fields must be passed in order to sample photons'

            # todo implement chunking for large datasets
            self.rebinned_cube = np.abs(np.sum(fields[:, -1], axis=2)) ** 2  # select detector plane and sum over objects
            self.rebinned_cube = self.rescale_cube(self.rebinned_cube)  # interpolate onto pixel spacing

            if sp.quick_detect and self.product == 'rebinned_cube':
                num_fake_events = int(ap.star_flux * sp.sample_time * np.sum(self.rebinned_cube))

                if sp.verbose:
                    print(f"star flux: {ap.star_flux}, cube sum: {np.sum(self.rebinned_cube)}, num fake events: {num_fake_events}")

                self.rebinned_cube *= num_fake_events
                self.photons = None
            else:
                max_steps = self.max_chunk(self.rebinned_cube)
                num_chunks = int(np.ceil(len(self.rebinned_cube)/max_steps))
                dprint(len(self.rebinned_cube), max_steps, len(self.rebinned_cube)/max_steps, num_chunks)
                self.photons = np.empty((4,0))
                for ic in range(num_chunks):
                    self.photons = self.get_photons(self.rebinned_cube[ic*max_steps:(ic+1)*max_steps],
                                                    chunk_step=abs_step + ic*max_steps)
                    if sp.degrade_photons:
                        self.photons = self.degrade_photons(self.photons)

                    if self.product == 'photons':
                        self.save_photontable(photonlist=self.photons, index=None, populate_subsidiaries=False)

                    elif self.product == 'rebinned_cube':
                        self.rebinned_cube[ic*max_steps:(ic+1)*max_steps] = self.rebin_list(self.photons, time_inds=[ic*max_steps,(ic+1)*max_steps])

                # only give the option to index and add Header after all MKID chunks are completed and even then it can be overwritten
                # for the opportunity to have chunked fields
                if finalise_photontable and self.product == 'photons':
                    self.save_photontable(photonlist=[], index=('ultralight', 6), populate_subsidiaries=True)
                    self.photontable_exists = True

            self.save_instance()

        return {'photons': self.photons, 'rebinned_cube': self.rebinned_cube}

    def quantize(self, fields, abs_step=0):
        self.rebinned_cube = np.abs(np.sum(fields[:, -1], axis=2)) ** 2  # select detector plane and sum over objects
        self.rebinned_cube = self.rescale_cube(self.rebinned_cube)  # interpolate onto pixel spacing

        if sp.quick_detect and self.product == 'rebinned_cube':
            num_fake_events = int(ap.star_flux * sp.sample_time * np.sum(self.rebinned_cube))

            if sp.verbose:
                print(f"star flux: {ap.star_flux}, cube sum: {np.sum(self.rebinned_cube)}, num fake events: {num_fake_events}")

            self.rebinned_cube *= num_fake_events
            self.photons = None

        else:
            max_steps = self.max_chunk(self.rebinned_cube)
            num_chunks = int(np.ceil(len(self.rebinned_cube) / max_steps))
            # dprint(len(self.rebinned_cube), max_steps, len(self.rebinned_cube) / max_steps, num_chunks)
            dprint(f"@Rup Please specify what you are trying to print here")
            #TODO Rupert-put a helpful print statement here instead
            self.photons = np.empty((4, 0))
            if self.product == 'photons':
                if num_chunks == 1:
                    self.photons = self.get_photons(self.rebinned_cube)
                    if self.usesave: self.save_photontable(photonlist=self.photons, index=('ultralight', 6), populate_subsidiaries=True)
                else:
                    for ic in range(num_chunks):
                        cspan = (ic * max_steps, (ic + 1) * max_steps)
                        self.photons = self.get_photons(self.rebinned_cube[cspan[0]: cspan[1]], chunk_step=abs_step + cspan[0])
                        if self.usesave: self.save_photontable(photonlist=self.photons, index=None, populate_subsidiaries=False)

                    # only give the option to index and add Header after all MKID chunks are completed and even then it
                    # can be overwritten for the opportunity to have chunked fields
                    if self.usesave:
                        self.save_photontable(photonlist=[], index=('ultralight', 6), populate_subsidiaries=True)
                        self.photontable_exists = True

                self.save_instance()
                return {'photons': self.photons}

            elif self.product == 'rebinned_cube':
                if num_chunks == 1:
                    self.photons = self.get_photons(self.rebinned_cube)
                    self.rebinned_cube = self.rebin_list(self.photons)
                else:
                    for ic in range(num_chunks):
                        cspan = (ic * max_steps, (ic + 1) * max_steps)
                        self.photons = self.get_photons(self.rebinned_cube[cspan[0]: cspan[1]], chunk_step=abs_step + cspan[0])
                        self.rebinned_cube[cspan[0]: cspan[1]] = self.rebin_list(self.photons, time_inds=[cspan[0], cspan[1]])

                self.save_instance()
                return {'rebinned_cube': self.rebinned_cube}

    def __call__(self, fields=None, abs_step=0):
        """
        Generate observation data with Camera

        Parameters
        ----------
        separate_objects: bool
            Generate photons from all targets at once or generate separate lists for each object. Separate lists are
            neccesary for training PCD algorithm
        """
        if self.photontable_exists and self.usesave:
            return self.load_photontable()

        else:
            return self.quantize(fields, abs_step)

    def max_chunk(self, datacube, round_chunk=True):
        """
        Determines the maximum duration each chunk can be to fit within the memory limit

        :return: integer
        """

        # max_counts = 50e6
        num_events = self.num_from_cube(datacube)
        self.photons_size = num_events * 32  # in Bytes (verified using photons[:,0].nbytes)

        max_chunk = sp.memory_limit*1e9 / self.photons_size
        # max_chunk = 1
        print(f'This observation will produce {num_events} photons, which is {self.photons_size/1e9} GB, meaning no '
              f'more than {max_chunk} of this observation can fit in the memory at one time')
        max_chunk = min([len(datacube)*max_chunk, len(datacube)])

        if round_chunk:
            # get nice max cut
            round_nums = [1, 2, 5]
            nice_cut = 0
            i = 0
            while nice_cut <= max_chunk:
                nice_cut = round_nums[i % len(round_nums)] * 10 ** (i // len(round_nums))  # [1,2,5,10,20,50,100,200,...]
                i += 1

            max_chunk = round_nums[(i - 2) % len(round_nums)] * 10 ** ((i - 2) // len(round_nums))  # we want the nice value thats below max_chunk so go back 2

        print(f'Input cubes will be split up into time length {max_chunk}')

        return max_chunk

    def save_photontable(self, photonlist=[], index=('ultralight', 6), timesort=False, chunkshape=None, shuffle=True, bitshuffle=False,
                        ndx_shuffle=True, ndx_bitshuffle=False, populate_subsidiaries=True):
        """
        save the photonlist in the MKIDPipeline format (https://github.com/MazinLab/MKIDPipeline)

        This code is ported from build_pytables() https://github.com/MazinLab/MKIDPipeline/blob/develop/mkidpipeline/hdf/bin2hdf.py#L64
        commit 292dec0f5140f5f1f941cc482c2fdcd2dd223011

        """
        from mkidcore.config import yaml
        from mkidcore.corelog import getLogger
        from mkidcore.headers import ObsFileCols, ObsHeader
        import mkidcore
        from mkidcore import pixelflags
        from io import StringIO
        import tables

        beammap = np.arange(mp.array_size[0]*mp.array_size[1]).reshape(mp.array_size)
        flagmap = np.zeros_like(beammap)

        h5file = tables.open_file(iop.photonlist, mode="a", title="MKID Photon File")
        filter = tables.Filters(complevel=1, complib='blosc:lz4', shuffle=shuffle, bitshuffle=bitshuffle,
                                fletcher32=False)

        if "/Photons" not in h5file:
            group = h5file.create_group("/", 'Photons', 'Photon Information')
            table = h5file.create_table(group, name='PhotonTable', description=ObsFileCols, title="Photon Datatable",
                                        expectedrows=len(photonlist), filters=filter, chunkshape=chunkshape)
        else:
            table = h5file.root.Photons.PhotonTable

        if photonlist != []:
            xs = np.int_(photonlist[2])
            ys = np.int_(photonlist[3])
    
            ResID = beammap[xs, ys]
            photons = np.zeros(len(photonlist[0]),
                               dtype=np.dtype([('ResID', np.uint32), ('Time', np.uint32), ('Wavelength', np.float32),
                                               ('SpecWeight', np.float32), ('NoiseWeight', np.float32)]))
    
            photons['ResID'] = ResID
            photons['Time'] = photonlist[0] * 1e6 # seconds -> microseconds
            photons['Wavelength'] = self.wave_cal(photonlist[1])
            if timesort:
                photons.sort(order=('Time', 'ResID'))
                getLogger(__name__).warning('Sorting photon data on time for {}'.format(iop.photonlist))
            elif not np.all(photons['ResID'][:-1] <= photons['ResID'][1:]):
                getLogger(__name__).warning('binprocessor.extract returned data that was not sorted on ResID, sorting'
                                            '({})'.format(iop.photonlist))
                photons.sort(order=('ResID', 'Time'))


            table.append(photons)
    
            getLogger(__name__).debug('Table Populated for {}'.format(iop.photonlist))
            
        if index:
            index_filter = tables.Filters(complevel=1, complib='blosc:lz4', shuffle=ndx_shuffle,
                                          bitshuffle=ndx_bitshuffle,
                                          fletcher32=False)

            def indexer(col, index, filter=None):
                if isinstance(index, bool):
                    col.create_csindex(filters=filter)
                else:
                    col.create_index(optlevel=index[1], kind=index[0], filters=filter)

            indexer(table.cols.Time, index, filter=index_filter)
            getLogger(__name__).debug('Time Indexed for {}'.format(iop.photonlist))
            indexer(table.cols.ResID, index, filter=index_filter)
            getLogger(__name__).debug('ResID Indexed for {}'.format(iop.photonlist))
            indexer(table.cols.Wavelength, index, filter=index_filter)
            getLogger(__name__).debug('Wavelength indexed for {}'.format(iop.photonlist))
            getLogger(__name__).debug('Table indexed ({}) for {}'.format(index, iop.photonlist))
            print('Table indexed ({}) for {}'.format(index, iop.photonlist))
        else:
            getLogger(__name__).debug('Skipping Index Generation for {}'.format(iop.photonlist))

        if populate_subsidiaries:
            group = h5file.create_group("/", 'BeamMap', 'Beammap Information', filters=filter)
            h5file.create_array(group, 'Map', beammap, 'resID map')
            h5file.create_array(group, 'Flag', flagmap, 'flag map')
            getLogger(__name__).debug('Beammap Attached to {}'.format(iop.photonlist))

            h5file.create_group('/', 'header', 'Header')
            headerTable = h5file.create_table('/header', 'header', ObsHeader, 'Header')
            headerContents = headerTable.row
            headerContents['isWvlCalibrated'] = True
            headerContents['isFlatCalibrated'] = True
            headerContents['isSpecCalibrated'] = True
            headerContents['isLinearityCorrected'] = True
            headerContents['isPhaseNoiseCorrected'] = True
            headerContents['isPhotonTailCorrected'] = True
            headerContents['timeMaskExists'] = False
            headerContents['startTime'] = int(time.time())
            headerContents['expTime'] = np.ceil(sp.sample_time * sp.numframes)
            headerContents['wvlBinStart'] = ap.wvl_range[0] * 1e9
            headerContents['wvlBinEnd'] = ap.wvl_range[1] * 1e9
            headerContents['energyBinWidth'] = 0.1  #todo check this
            headerContents['target'] = ''
            headerContents['dataDir'] = iop.testdir
            headerContents['beammapFile'] = ''
            headerContents['wvlCalFile'] = ''
            headerContents['fltCalFile'] = ''
            headerContents['metadata'] = ''

            out = StringIO()
            yaml.dump({'flags': mkidcore.pixelflags.FLAG_LIST}, out)
            out = out.getvalue().encode()
            if len(out) > mkidcore.headers.METADATA_BLOCK_BYTES:  # this should match mkidcore.headers.ObsHeader.metadata
                raise ValueError("Too much metadata! {} KB needed, {} allocated".format(len(out) // 1024,
                                                                                        mkidcore.headers.METADATA_BLOCK_BYTES // 1024))
            headerContents['metadata'] = out

            headerContents.append()
            getLogger(__name__).debug('Header Attached to {}'.format(iop.photonlist))

        h5file.close()
        getLogger(__name__).debug('Done with {}'.format(iop.photonlist))

        # self.photontable_exists = True

    def load_photontable(self):
        """Load photon list from pipeline's photon table h5"""
        import tables
        h5file = tables.open_file(iop.photonlist, "r")
        table = h5file.root.Photons.PhotonTable
        self.is_wave_cal = h5file.root.header.header.description.isWvlCalibrated

        self.photons = np.zeros((4, len(table)))
        self.photons[0] = table[:]['Time'] /1e6
        self.photons[1] = table[:]['Wavelength']

        rev_beam = []
        [[rev_beam.append([c,r]) for r in range(mp.array_size[1])] for c in range(mp.array_size[0])]
        rev_beam = np.array(rev_beam)
        self.photons[2:] = (rev_beam[table[:]['ResID']].T)
        self.rebinned_cube = None
        print(f"Loaded photon list from table at{h5file}")

        return {'photons': self.photons}

    def save_instance(self):
        """
        Save the whole Camera instance

        This is a defensive way to write pickle.write, allowing for very large files on all platforms
        """
        max_bytes = 2 ** 31 - 1
        bytes_out = pickle.dumps(self)
        n_bytes = sys.getsizeof(bytes_out)
        with open(self.name, 'wb') as f_out:
            for idx in range(0, n_bytes, max_bytes):
                f_out.write(bytes_out[idx:idx + max_bytes])

    def load_instance(self):
        """
        This is a defensive way to write pickle.load, allowing for very large files on all platforms
        """
        max_bytes = 2 ** 31 - 1
        try:
            input_size = os.path.getsize(self.name)
            bytes_in = bytearray(0)
            with open(self.name, 'rb') as f_in:
                for _ in range(0, input_size, max_bytes):
                    bytes_in += f_in.read(max_bytes)
            obj = pickle.loads(bytes_in)
        except:
            return None
        return obj

    def num_from_cube(self, datacube):
        return int(ap.star_flux * sp.sample_time * np.sum(datacube))

    def get_photons(self, datacube, chunk_step=0, plot=False):
        """
        Given an intensity spectralcube and timestep create a quantized photon list

        :param datacube:
        :param step:
        :param plot:
        :return:
        """

        if mp.QE_var:
            datacube *= self.QE_map.T

        num_events = self.num_from_cube(datacube)

        if sp.verbose: print(f"star flux: {ap.star_flux}, cube sum: {np.sum(datacube)}, numframes: {self.numframes},"
              f"num events: {num_events:e}")

        photons = self.sample_cube(datacube, num_events)
        photons[0] += chunk_step
        photons[0] = self.assign_time(photons[0])
        photons[1] = self.assign_phase(photons[1])

        if plot:
            grid(self.rebin_list(photons), title='get_photons')

        if sp.degrade_photons:
            photons = self.degrade_photons(photons)

        return photons

    def degrade_photons(self, photons, plot=False):
        if plot:
            grid(self.rebin_list(photons), title='before degrade')

        if mp.dark_counts:
            dark_photons = self.get_bad_packets(type='dark')
            dprint(photons.shape, dark_photons.shape, 'dark')
            photons = np.hstack((photons, dark_photons))

        if mp.hot_pix:
            hot_photons = self.get_bad_packets(type='hot')
            photons = np.hstack((photons, hot_photons))
            # stem = add_hot(stem)

        if plot:
            grid(self.rebin_list(photons), title='after bad')

        if mp.phase_uncertainty:
            photons[1] *= self.responsivity_error_map[np.int_(photons[2]), np.int_(photons[3])]
            photons, idx = self.apply_phase_offset_array(photons, self.sigs)

        # thresh =  photons[1] < self.basesDeg[np.int_(photons[3]),np.int_(photons[2])]
        if mp.phase_background:
            thresh =  -photons[1] > 3*self.sigs[-1,np.int_(photons[3]), np.int_(photons[2])]
            photons = photons[:, thresh]

        if mp.remove_close:
            stem = self.arange_into_stem(photons.T, (self.array_size[0], self.array_size[1]))
            stem = self.remove_close(stem)
            photons = self.ungroup(stem)

        if plot:
            grid(self.rebin_list(photons), title='after remove close')

        # This step was taking a long time
        # stem = arange_into_stem(photons.T, (self.array_size[0], self.array_size[1]))
        # cube = make_datacube(stem, (self.array_size[0], self.array_size[1], ap.n_wvl_final))
        # # ax7.imshow(cube[0], origin='lower', norm=LogNorm(), cmap='inferno', vmin=1)
        # cube /= self.QE_map
        # photons = ungroup(stem)

        return photons

    def arange_into_cube(self, packets, size):
        # print 'Sorting packets into xy grid (no phase or time sorting)'
        cube = [[[] for i in range(size[0])] for j in range(size[1])]
        for ip, p in enumerate(packets):
            x = np.int_(p[2])
            y = np.int_(p[3])
            cube[x][y].append([p[0] ,p[1]])
            if len(packets) >= 1e7 and ip % 1000:
                misc.progressBar(value=ip, endvalue=len(packets))
        # print cube[x][y]
        # cube = time_sort(cube)
        return cube

    def responvisity_scaling_map(self, plot=False):
        """Assigns each pixel a phase responsivity between 0 and 1"""
        dist = Distribution(gaussian(mp.r_mean, mp.r_sig, np.linspace(0, 2, mp.res_elements)), interpolation=True)
        responsivity = dist(self.array_size[0] * self.array_size[1])[0]/float(mp.res_elements) * 2
        if plot:
            plt.xlabel('Responsivity')
            plt.ylabel('#')
            plt.hist(responsivity)
            plt.show()
        responsivity = np.reshape(responsivity, self.array_size)
        if plot:
            quick2D(responsivity)#plt.imshow(QE)
            # plt.show()

        return responsivity

    def array_QE(self, plot=False):
        """Assigns each pixel a phase responsivity between 0 and 1"""
        dist = Distribution(gaussian(mp.g_mean, mp.g_sig, np.linspace(0, 1, mp.res_elements)), interpolation=True)
        QE = dist(self.array_size[0] * self.array_size[1])[0]/float(mp.res_elements)
        if plot:
            plt.xlabel('Responsivity')
            plt.ylabel('#')
            plt.hist(QE)
            plt.show()
        QE = np.reshape(QE, self.array_size[::-1])
        if plot:
            quick2D(QE)#plt.imshow(QE)
            # plt.show()

        return QE

    def assign_spectral_res(self, plot=False):
        """Assigning each pixel a spectral resolution (at 800nm)"""
        dist = Distribution(gaussian(0.5, 0.25, np.linspace(-0.2, 1.2, mp.res_elements)), interpolation=True)
        # print(f"Mean R = {mp.R_mean}")
        Rs = (dist(self.array_size[0]*self.array_size[1])[0]/float(mp.res_elements)-0.5)*mp.R_sig + mp.R_mean#
        if plot:
            plt.xlabel('R')
            plt.ylabel('#')
            plt.hist(Rs)
            plt.show()
        Rs = np.reshape(Rs, self.array_size)

        if plot:
            plt.figure()
            plt.imshow(Rs)
            plt.show()
        return Rs

    def get_R_hyper(self, Rs, plot=False):
        """Each pixel of the array has a matrix of probabilities that depends on the input wavelength"""
        print('Creating a cube of R standard deviations')
        m = (mp.R_spec*Rs)/(ap.wvl_range[1] - ap.wvl_range[0]) # looses R of 10% over the 700 wvl_range
        c = Rs-m*ap.wvl_range[0] # each c depends on the R @ 800
        waves = np.ones((np.shape(m)[1],np.shape(m)[0],ap.n_wvl_final+5))*np.linspace(ap.wvl_range[0],ap.wvl_range[1],ap.n_wvl_final+5)
        waves = np.transpose(waves) # make a tensor of 128x128x10 where every 10 vector is 800... 1500
        R_spec = m * waves + c # 128x128x10 tensor is now lots of simple linear lines e.g. 50,49,.. 45
        # probs = np.ones((np.shape(R_spec)[0],np.shape(R_spec)[1],np.shape(R_spec)[2],
        #                 mp.res_elements))*np.linspace(0, 1, mp.res_elements)
        #                         # similar to waves but 0... 1 using 128 elements
        # R_spec = np.repeat(R_spec[:,:,:,np.newaxis], mp.res_elements, 3) # creat 128 repeats of R_spec so (10,128,128,128)
        # mp.R_probs = gaussian(0.5, R_spec, probs) #each xylocation is gaussian that gets wider for longer wavelengths
        sigs_w = (waves/R_spec)/2.35 #R = w/dw & FWHM = 2.35*sig

        # plt.plot(range(0,1500),spec.phase_cal(np.arange(0,1500)))
        # plt.show()
        sigs_p = self.phase_cal(sigs_w) - self.phase_cal(np.zeros_like((sigs_w)))

        if plot:
            plt.figure()
            plt.plot(R_spec[:, 0, 0])
            plt.plot(R_spec[:,50,15])
            # plt.figure()
            # plt.plot(sigs_w[:,0,0])
            # plt.plot(sigs_w[:,50,15])
            # plt.figure()
            # plt.plot(sigs_p[:, 0, 0])
            # plt.plot(sigs_p[:, 50, 15])
            # plt.figure()
            # plt.imshow(sigs_p[:,:,0], aspect='auto')
            view_spectra(sigs_w, show=False)
            view_spectra(sigs_p, show=False)
            plt.figure()
            plt.plot(np.mean(sigs_w, axis=(1,2)))
            plt.figure()
            plt.plot(np.mean(sigs_p, axis=(1,2)))
            # plt.imshow(mp.R_probs[:,0,0,:])
            plt.show(block=True)
        return sigs_p

    def apply_phase_scaling(photons, ):
        """
        From things like resonator Q, bias power, quasiparticle losses

        :param photons:
        :return:
        """

    def apply_phase_offset_array(self, photons, sigs):
        """
        From things like IQ phase offset noise

        :param photons:
        :param sigs:
        :return:
        """
        wavelength = self.wave_cal(photons[1])

        idx = self.wave_idx(wavelength)

        bad = np.where(np.logical_or(idx>=len(sigs), idx<0))[0]

        photons = np.delete(photons, bad, axis=1)
        idx = np.delete(idx, bad)

        distortion = np.random.normal(np.zeros((photons[1].shape[0])),
                                      sigs[idx,np.int_(photons[2]), np.int_(photons[3])])

        photons[1] += distortion

        return photons, idx

    def apply_phase_distort(self, phase, loc, sigs):
        """
        Simulates phase height of a real detector system per photon
        proper will spit out the true phase of each photon it propagates. this function will give it
        a 'measured' phase based on the resolution of the detector, and a Gaussian distribution around
        the center of the resolution bin

        :param phase: real(exact) phase information from Proper
        :param loc:
        :param sigs:
        :return: distorted phase
        """
        # phase = phase + mp.phase_distortions[ip]
        wavelength = self.wave_cal(phase)
        idx = self.wave_idx(wavelength)

        if phase != 0 and idx<len(sigs):
            phase = np.random.normal(phase,sigs[idx,loc[0],loc[1]],1)[0]
        return phase

    def wave_idx(self, wavelength):
        m = float(ap.n_wvl_final-1)/(ap.wvl_range[1] - ap.wvl_range[0])
        c = -m*ap.wvl_range[0]
        idx = wavelength*m + c
        # return np.int_(idx)
        return np.int_(np.round(idx))

    def phase_cal(self, wavelengths):
        '''Wavelength in nm'''
        phase = mp.wavecal_coeffs[0]*wavelengths + mp.wavecal_coeffs[1]
        return phase

    def wave_cal(self, phase):
        wave = (phase - mp.wavecal_coeffs[1])/(mp.wavecal_coeffs[0])
        return wave

    def assign_phase_background(self, plot=False):
        """assigns each pixel a baseline phase"""
        dist = Distribution(gaussian(0.5, 0.25, np.linspace(-0.2, 1.2, mp.res_elements)), interpolation=True)

        basesDeg = dist(self.array_size[0]*self.array_size[1])[0]/float(mp.res_elements)*mp.bg_mean/mp.g_mean
        if plot:
            plt.xlabel('basesDeg')
            plt.ylabel('#')
            plt.title('Background Phase')
            plt.hist(basesDeg)
            plt.show(block=True)
        basesDeg = np.reshape(basesDeg, self.array_size)
        if plot:
            plt.title('Background Phase--Reshaped')
            plt.imshow(basesDeg)
            plt.show(block=True)
        return basesDeg


    def create_bad_pix(self, QE_map, plot=False):
        amount = int(self.array_size[0]*self.array_size[1]*(1.-mp.pix_yield))

        bad_ind = random.sample(list(range(self.array_size[0]*self.array_size[1])), amount)

        dprint(f"Bad indices = {len(bad_ind)}, # MKID pix = { self.array_size[0]*self.array_size[1]}, "
               f"Pixel Yield = {mp.pix_yield}, amount? = {amount}")

        # bad_y = random.sample(y, amount)
        bad_y = np.int_(np.floor(bad_ind/self.array_size[1]))
        bad_x = bad_ind % self.array_size[1]

        # print(f"responsivity shape  = {responsivities.shape}")
        QE_map = np.array(QE_map)

        QE_map[bad_x, bad_y] = 0

        if plot:
            plt.xlabel('x pix')
            plt.ylabel('y pix')
            plt.title('Bad Pixel Map')
            plt.imshow(QE_map)
            cax = plt.colorbar()
            cax.set_label('Responsivity')
            plt.show()

        return QE_map


    def create_bad_pix_center(self, responsivities):
        res_elements=self.array_size[0]
        # responsivities = np.zeros()
        for x in range(self.array_size[1]):
            dist = Distribution(gaussian(0.5, 0.25, np.linspace(0, 1, mp.res_elements)), interpolation=False)
            dist = np.int_(dist(int(self.array_size[0]*mp.pix_yield))[0])#/float(mp.res_elements)*np.int_(self.array_size[0]) / mp.g_mean)
            # plt.plot(dist)
            # plt.show()
            dead_ind = []
            [dead_ind.append(el) for el in range(self.array_size[0]) if el not in dist]
            responsivities[x][dead_ind] = 0

        return responsivities

    def get_bad_packets(self, type='dark'):
        if type == 'hot':
            n_device_counts = self.total_hot
        elif type == 'dark':
            n_device_counts = self.total_dark
            dprint(self.total_dark, mp.dark_bright, sp.sample_time)
        else:
            print("type currently has to be 'hot' or 'dark'")
            raise AttributeError

        if n_device_counts % 1 > np.random.uniform(0,1,1):
            n_device_counts += 1

        n_device_counts = int(n_device_counts)
        photons = np.zeros((4, n_device_counts))
        if n_device_counts > 0:
            if type == 'hot':
                phases = np.random.uniform(-120, 0, n_device_counts)
                hot_ind = np.random.choice(range(len(self.hot_locs[0])), n_device_counts)
                bad_pix = self.hot_locs[:, hot_ind]
            elif type == 'dark':
                dist = Distribution(gaussian(0, 0.25, np.linspace(0, 1, mp.res_elements)), interpolation=False)
                phases = dist(n_device_counts)[0]
                max_phase = max(phases)
                phases = -phases*120/max_phase
                bad_pix_options = self.create_false_pix(self.dark_pix_frac * self.array_size[0] * self.array_size[1])
                bad_ind = np.random.choice(range(len(bad_pix_options[0])), n_device_counts)
                bad_pix = bad_pix_options[:, bad_ind]

            photons[0] = np.random.uniform(sp.startframe * sp.sample_time, sp.numframes * sp.sample_time, n_device_counts)
            photons[1] = phases
            photons[2:] = bad_pix

        return photons

    def create_false_pix(self, amount):
        # print(f"amount = {amount}")
        bad_ind = random.sample(list(range(self.array_size[0]*self.array_size[1])), int(amount))
        bad_y = np.int_(np.floor(bad_ind / self.array_size[1]))
        bad_x = bad_ind % self.array_size[1]

        return np.array([bad_x, bad_y])


    def remove_bad(self, frame, QE):
        bad_map = np.ones((sp.grid_size,sp.grid_size))
        bad_map[QE[:-1,:-1]==0] = 0
        # quick2D(QE, logZ =False)
        # quick2D(bad_map, logZ =False)
        frame = frame*bad_map
        return frame

    def sample_cube(self, datacube, num_events):
        # print 'creating photon data from reference cube'

        dist = Distribution(datacube, interpolation=2)

        photons = dist(num_events)
        return photons

    def assign_calibtime(self, photons, step):
        meantime = step*sp.sample_time
        # photons = photons.astype(float)#np.asarray(photons[0], dtype=np.float64)
        # photons[0] = photons[0] * ps.mp.frame_time
        timedist = np.random.uniform(meantime-sp.sample_time/2, meantime+sp.sample_time/2, len(photons[0]))
        photons = np.vstack((timedist, photons))
        return photons

    def assign_time(self, time_idx):
        return time_idx*sp.sample_time

    def assign_phase(self, phase_idx):
        """
        idx -> phase

        :parameter: idx (list)

        :return: phase (list)
        """
        phase_idx = np.array(phase_idx)
        c = ap.wvl_range[0]
        m = (ap.wvl_range[1] - ap.wvl_range[0])/ap.n_wvl_final
        wavelengths = phase_idx*m + c

        return self.phase_cal(wavelengths)

    def make_datacube_from_list(self, packets):
        phase_band = self.phase_cal(ap.wvl_range)
        bins = [np.linspace(phase_band[0], phase_band[1], ap.n_wvl_final + 1),
                range(self.array_size[0]+1),
                range(self.array_size[1]+1)]
        datacube, _ = np.histogramdd(packets[:,1:], bins)

        return datacube

    def rebin_list(self, photons, time_inds=None):
        phase_band = self.phase_cal(ap.wvl_range)
        if not time_inds:
            time_inds = [0, self.numframes]
        bins = [np.linspace(sp.sample_time * time_inds[0], sp.sample_time * time_inds[1], time_inds[1] - time_inds[0] + 1),
                np.linspace(phase_band[0], phase_band[1], ap.n_wvl_final + 1),
                range(self.array_size[0] + 1),
                range(self.array_size[1] + 1)]
        if self.is_wave_cal:
            bins[1] = self.wave_cal(np.linspace(phase_band[0], phase_band[1], ap.n_wvl_final + 1))
        rebinned_cube, _ = np.histogramdd(photons.T, bins)
        return rebinned_cube

    def cut_max_count(self, datacube):
        image = np.sum(datacube, axis=0)
        max_counts = self.max_count * sp.sample_time
        maxed = image > max_counts
        scaling_map = max_counts / image
        scaling_map[scaling_map > 1] = 1
        scaled_cube = datacube * scaling_map
        dprint(f'datacube went from {np.sum(datacube)} to {np.sum(scaled_cube)} '
               f'(factor of {np.sum(scaled_cube)/np.sum(datacube)}) during max count cut')

        return scaled_cube

    def remove_close(self, stem):
        print('removing close photons')
        for x in range(len(stem)):
            for y in range(len(stem[0])):
                # print(x,y)
                if len(stem[x][y]) > 1:
                    events = np.array(stem[x][y])
                    timesort = np.argsort(events[:, 0])
                    events = events[timesort]
                    detected = [0]
                    idx = 0
                    while idx != None:
                        these_times = events[idx:, 0] - events[idx, 0]
                        detect, _ = next(((i, v) for (i, v) in enumerate(these_times) if v > mp.dead_time),
                                         (None, None))
                        if detect != None:
                            detect += idx
                            detected.append(detect)
                        idx = detect

                    missed = [ele for ele in range(detected[-1] + 1) if ele not in detected]
                    events = np.delete(events, missed, axis=0)
                    stem[x][y] = events
                    # dprint(x, len(missed))
        return stem

    def arange_into_stem(self, packets, size):
        # print 'Sorting packets into xy grid (no phase or time sorting)'
        stem = [[[] for i in range(size[0])] for j in range(size[1])]
        # dprint(np.shape(cube))
        # plt.hist(packets[:,1], bins=100)
        # plt.show()
        for ip, p in enumerate(packets):
            x = np.int_(p[2])
            y = np.int_(p[3])
            stem[x][y].append([p[0], p[1]])
            if len(packets) >= 1e7 and ip % 10000 == 0: misc.progressBar(value=ip, endvalue=len(packets))
        # print cube[x][y]
        # cube = time_sort(cube)
        return stem

    def ungroup(self, stem):
        photons = np.empty((0, 4))
        for x in range(mp.array_size[1]):
            for y in range(mp.array_size[0]):
                # print(x, y)
                if len(stem[x][y]) > 0:
                    events = np.array(stem[x][y])
                    xy = [[x, y]] * len(events) if len(events.shape) == 2 else [x, y]
                    events = np.append(events, xy, axis=1)
                    photons = np.vstack((photons, events))
                    # print(events, np.shape(events))
                    # timesort = np.argsort(events[:, 0])
                    # events = events[timesort]
                    # sep = events[:, 0] - np.roll(events[:, 0], 1, 0)
        return photons.T

    def get_ideal_photons(self, cube, step):
        raise NotImplementedError
        ncounts = np.sum(cube)
        photons = np.zeros((ncounts, 4))
        photons[:,0] = step*sp.sample_time
        for iw, slice in enumerate(cube):
            nslicecounts = np.sum(cube)
            photons[:, iw*nslicecounts:(iw+1)*nslicecounts] = ap.wvl_samp[iw]
            for x in mp.array_size[0]:
                for y in mp.array_size[1]:
                    npixcounts = slice[x,y]
                    photons[:,:,x*npixcounts:(x+1)*npixcounts,y*npixcounts:(y+1)*npixcounts] = [x,y]

    def rescale_cube(self, rebinned_cube, conserve=True):
        if conserve:
            total = np.sum(rebinned_cube)
        nyq_sampling = ap.wvl_range[0]*360*3600/(2*np.pi*tp.entrance_d)
        self.sampling = nyq_sampling*sp.beam_ratio*2  # nyq sampling happens at sp.beam_ratio = 0.5
        x = np.arange(-sp.grid_size*self.sampling/2, sp.grid_size*self.sampling/2, self.sampling)
        xnew = np.arange(-self.array_size[0]*self.platescale/2, self.array_size[0]*self.platescale/2, self.platescale)
        ynew = np.arange(-self.array_size[1]*self.platescale/2, self.array_size[1]*self.platescale/2, self.platescale)
        mkid_cube = np.zeros((rebinned_cube.shape[0], rebinned_cube.shape[1], self.array_size[0], self.array_size[1]))
        for d, datacube in enumerate(rebinned_cube):
            for s, slice in enumerate(datacube):
                f = interpolate.interp2d(x, x, slice, kind='cubic')
                mkid_cube[d, s] = f(ynew, xnew)
        mkid_cube = mkid_cube*np.sum(datacube)/np.sum(mkid_cube)
        # grid(mkid_cube, logZ=True, show=True, extract_center=False, title='post')
        rebinned_cube = mkid_cube

        rebinned_cube[rebinned_cube < 0] *= -1

        if conserve:
            rebinned_cube *= total/np.sum(rebinned_cube)

        return rebinned_cube


if __name__ == '__main__':
    iop.update_testname('MKIDS_module_test')
    sp.quick_detect = True

    cam = Camera()
    observation = cam()
    print(observation.keys())
    grid(observation['rebinned_photons'], nstd=5)
